from __future__ import annotations

__all__ = (
    "Middleware",
    "chat_action",
)

import asyncio
from typing import TYPE_CHECKING
from typing import Any
from typing import Generic
from typing import TypeVar

if TYPE_CHECKING:
    from .enums import ChatAction
    from .types import Message
    from .typing import HandlerCallable
    from .typing import HandlerMiddleware
    from .typing import MiddlewareType

T = TypeVar("T")


class Middleware(Generic[T]):

    def __init__(
        self,
        obj: MiddlewareType[T],
        *,
        is_handler: bool = False,
        is_local: bool = False,
        **params: dict[str, Any],
    ) -> None:
        self.obj = obj
        self.is_handler = is_handler
        self.is_local = is_local
        self.params = params

    def __call__(self, obj: T) -> T:
        return self.obj(obj, **self.params)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Middleware) and (  # yapf: disable
            other is self or (
                other.obj is self.obj
                and other.is_handler is self.is_handler
                and other.is_local is self.is_local
            )
        )


def chat_action(action: ChatAction) -> HandlerMiddleware[Message]:
    """Send chat action and automatically update it if your operation takes
    more than 5 seconds to complete.

    Usage::

        @on_message(middleware=[chat_action(ChatAction.UPLOAD_PHOTO)])
        async def slow_operation(message: Message) -> None:
            ...

    See: :class:`ChatAction <yatbaf.enums.ChatAction>`

    :param action: Chat action.
    """

    def outer(handler: HandlerCallable[Message]) -> HandlerCallable[Message]:

        async def inner(update: Message) -> None:
            event = asyncio.Event()

            async def _task() -> None:
                while not event.is_set():
                    await update.bot.send_chat_action(update.chat.id, action)
                    await asyncio.sleep(4.5)  # update status every 4.5 sec

            update.bot.tasks.add(
                _task(),
                name=f"chat-action-{update.message_id}",
            )
            try:
                await handler(update)
            finally:
                event.set()

        return inner

    return outer



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Literal
from typing import final

from .abc import TelegramType

if TYPE_CHECKING:
    from .chat import Chat


@final
class InaccessibleMessage(TelegramType):
    """This object describes a message that was deleted or is otherwise
    inaccessible to the bot.

    See: https://core.telegram.org/bots/api#inaccessiblemessage
    """

    chat: Chat
    """Chat the message belonged to."""

    message_id: int
    """Unique message identifier inside the chat."""

    date: Literal[0] = 0
    """Always 0. The field can be used to differentiate regular and inaccessible
    messages.
    """



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import final

from yatbaf.types import Message

from .abc import TelegramMethodWithFile

if TYPE_CHECKING:
    from yatbaf.enums import ParseMode
    from yatbaf.types import MessageEntity
    from yatbaf.types import ReplyParameters
    from yatbaf.typing import InputFile
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt
    from yatbaf.typing import NoneStr
    from yatbaf.typing import ReplyMarkup


@final
class SendVideo(TelegramMethodWithFile[Message]):
    """See :meth:`yatbaf.bot.Bot.send_video`"""

    chat_id: str | int
    video: InputFile | str
    message_thread_id: NoneInt = None
    duration: NoneInt = None
    width: NoneInt = None
    height: NoneInt = None
    thumbnail: InputFile | str | None = None
    caption: NoneStr = None
    parse_mode: ParseMode | None = None
    caption_entities: list[MessageEntity] | None = None
    has_spoiler: NoneBool = None
    supports_streaming: NoneBool = None
    disable_notification: NoneBool = None
    protect_content: NoneBool = None
    reply_parameters: ReplyParameters | None = None
    reply_markup: ReplyMarkup | None = None

    __meth_file_fields__: ClassVar[tuple[str, ...]] = (
        "video",
        "thumbnail",
    )



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneInt

from .abc import TelegramType


@final
class ResponseParameters(TelegramType):
    """Describes why a request was unsuccessful.

    See: https://core.telegram.org/bots/api#responseparameters
    """

    migrate_to_chat_id: NoneInt = None
    """*Optional.* The group has been migrated to a supergroup with the
    specified identifier.
    """

    retry_after: NoneInt = None
    """*Optional.* In case of exceeding flood control, the number of seconds
    left to wait before the request can be repeated.
    """



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneInt

from .abc import TelegramType
from .photo_size import PhotoSize


@final
class VideoNote(TelegramType):
    """This object represents a video message.

    See: https://core.telegram.org/bots/api#videonote
    """

    file_id: str
    """Identifier for this file, which can be used to download or reuse the file."""  # noqa: E501

    file_unique_id: str
    """Unique identifier for this file, which is supposed to be the same over
    time and for different bots.

    .. warning::

        Can't be used to download or reuse the file.
    """

    length: int
    """Video width and height (diameter of the video message) as defined by sender."""  # noqa: E501

    duration: int
    """Duration of the video in seconds as defined by sender."""

    thumbnail: PhotoSize | None = None
    """*Optional.* Video thumbnail."""

    file_size: NoneInt = None
    """*Optional.* File size in bytes."""



from __future__ import annotations

from typing import final

from .abc import TelegramMethod


@final
class SetChatStickerSet(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.set_chat_sticker_set`"""

    chat_id: str | int
    sticker_set_name: str



import re

import pytest

from yatbaf.filters import Text


def test_no_params():
    with pytest.raises(ValueError):
        Text()


def test_is_text(message):
    message.text = None
    assert not Text(startswith="foo").check(message)


@pytest.mark.parametrize("f,ic", (("FOO", False), ("foo", True)))
def test_startswith(message, f, ic):
    message.text = "FOO bar"
    assert Text(startswith=f, ignore_case=ic).check(message)


@pytest.mark.parametrize("f,ic", (("BAR", False), ("bar", True)))
def test_endswith(message, f, ic):
    message.text = "foo BAR"
    assert Text(endswith=f, ignore_case=ic).check(message)


def test_start_end(message):
    message.text = "foo baz bar"
    assert Text(startswith="foo", endswith="bar").check(message)


def test_start_end_any(message):
    message.text = "foo bar baz"
    assert (Text(startswith="foo", endswith="bar", any_=True).check(message))


@pytest.mark.parametrize("m,ic", (("Foo Bar", False), ("foo bar", True)))
def test_match(message, m, ic):
    message.text = "Foo Bar"
    assert Text(match=m, ignore_case=ic).check(message)


def test_match_false(message):
    message.text = "foo bar"
    assert not Text(match="foo b").check(message)


@pytest.mark.parametrize("c", ("bar", ["oof", "foo"]))
def test_contains(message, c):
    message.text = "foo bar baz"
    assert Text(contains=c).check(message)


def test_contains_false(message):
    message.text = "foo bar baz"
    assert not Text(contains="foobar").check(message)


@pytest.mark.parametrize("t", ("foobarbaz", "foo bar baz"))
def test_start_end_contains(message, t):
    message.text = t
    assert Text(
        startswith="fo",
        endswith="az",
        contains="bar",
    ).check(message)


@pytest.mark.parametrize("exp", (".+baz.*$", re.compile("^f.+az$")))
def test_regexp(message, exp):
    message.text = "foo bar baz"
    assert Text(regexp=exp).check(message)



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneBool
from yatbaf.typing import NoneStr

from .abc import TelegramType


@final
class LoginUrl(TelegramType):
    """This object represents a parameter of the inline keyboard button used to
    automatically authorize a user.

    See: https://core.telegram.org/bots/api#loginurl
    """

    url: str
    """An HTTPS URL to be opened with user authorization data added to the query
    string when the button is pressed. If the user refuses to provide
    authorization data, the original URL without information about the user will
    be opened. The data added is the same as described in `Receiving
    authorization data`_.

    .. _Receiving authorization data: https://core.telegram.org/widgets/login#receiving-authorization-data
    """  # noqa: E501

    forward_text: NoneStr = None
    """*Optional.* New text of the button in forwarded messages."""

    bot_username: NoneStr = None
    """*Optional.* Username of a bot, which will be used for user authorization.
    See `Setting up a bot`_ for more details. If not specified, the current
    bot's username will be assumed. The ``url``'s domain must be the same as the
    domain linked with the bot. See `Linking your domain to the bot`_ for more details.

    .. _Setting up a bot: https://core.telegram.org/widgets/login#setting-up-a-bot
    .. _Linking your domain to the bot: https://core.telegram.org/widgets/login#linking-your-domain-to-the-bot
    """  # noqa: E501

    request_write_access: NoneBool = None
    """*Optional.* Pass ``True`` to request the permission for your bot to send
    messages to the user.
    """



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from msgspec import field

from yatbaf.typing import NoneBool
from yatbaf.typing import NoneInt
from yatbaf.typing import NoneStr

from .abc import TelegramType
from .inaccessible_message import InaccessibleMessage
from .message import Message
from .user import User

if TYPE_CHECKING:
    from .maybe_inaccessible_message import MaybeInaccessibleMessage


@final
class CallbackQuery(TelegramType, kw_only=True):
    """This object represents an incoming callback query from a callback button
    in an inline keyboard. If the button that originated the query was attached
    to a message sent by the bot, the field ``message`` will be present. If the
    button was attached to a message sent via the bot (in inline mode), the
    field ``inline_message_id`` will be present. Exactly one of the fields data
    or ``game_short_name`` will be present.

    See: https://core.telegram.org/bots/api#callbackquery
    """

    id: str
    """Unique identifier for this query"""

    from_: User = field(name="from")
    """Sender"""

    if TYPE_CHECKING:
        message: MaybeInaccessibleMessage | None = None
        """*Optional.* Message with the callback button that originated the
        query. Note that message content and message date will not be available
        if the message is too old.
        """
    else:
        message: Message | None = None

    inline_message_id: NoneStr = None
    """*Optional.* Identifier of the message sent via the bot in inline mode,
    that originated the query.
    """

    chat_instance: str
    """Global identifier, uniquely corresponding to the chat to which the
    message with the callback button was sent. Useful for high scores in games.
    """

    data: NoneStr = None
    """*Optional.* Data associated with the callback button. Be aware that the
    message originated the query can contain no callback buttons with this data.
    """

    game_short_name: NoneStr = None
    """*Optional.* Short name of a `Game`_ to be returned, serves as the unique
    identifier for the game.

    .. _game: https://core.telegram.org/bots/api#games
    """

    def __post_init__(self) -> None:
        super().__post_init__()
        if (m := self.message) is not None and m.date == 0:
            self.message = InaccessibleMessage(
                chat=m.chat,
                message_id=m.message_id,
            )

    async def answer(
        self,
        text: NoneStr = None,
        show_alert: NoneBool = None,
        url: NoneStr = None,
        cache_time: NoneInt = None,
    ) -> bool:
        """Use this method to send answer to callback.

        See: :meth:`Bot.answer_callback_query <yatbaf.bot.Bot.answer_callback_query>`
        """  # noqa: E501

        return await self.bot.answer_callback_query(
            callback_query_id=self.id,
            text=text,
            show_alert=show_alert,
            url=url,
            cache_time=cache_time,
        )



from __future__ import annotations

__all__ = ("TelegramType",)

from typing import TYPE_CHECKING
from typing import Any
from typing import ClassVar
from typing import cast

from msgspec import Struct

if TYPE_CHECKING:
    from yatbaf.bot import Bot


class _usrctx:  # noqa: N801
    """:meta private:"""

    __slots__ = ("__usrctx__",)


class TelegramType(_usrctx, Struct, omit_defaults=True):
    """Base class for Telegram types"""

    __usrctx__: ClassVar[dict[str, Any]]
    __type_file_fields__: ClassVar[tuple[str, ...]] = ()

    def __post_init__(self) -> None:  # pragma: no cover
        self.__usrctx__ = {}  # type: ignore[misc]

    def __str__(self) -> str:  # pragma: no cover
        return self.__class__.__name__.lower()

    @property
    def ctx(self) -> dict[str, Any]:
        """Dict object.

        Use it to share data between middleware/handler in request context::

            @router.guard
            async def guard(update):
                update.ctx["foo"] = "bar"
                return True

            @router.middleware
            def middleware(handler):
                async def wrapper(update):
                    update.ctx["foo"] += "baz"
                    await handler(update)
                return wrapper

            @router
            async def handler(update):
                assert update.ctx["foo"] == "barbaz"
        """

        return self.__usrctx__

    @property
    def bot(self) -> Bot:
        """Bot instance.

        Use it to get access to :class:`Bot <yatbaf.bot.Bot>` instance inside
        handler function::

            @on_message
            async def handler(message: Message) -> None:
                await message.bot.leave_chat(message.chat.id)
        """

        try:
            return cast("Bot", self.__usrctx__["__bot_obj__"])
        except KeyError:
            raise RuntimeError("Bot not bound to this instance.") from None

    def _bind_bot_obj(self, bot: Bot) -> None:
        self.__usrctx__["__bot_obj__"] = bot

        for field_name in self.__struct_fields__:
            field = getattr(self, field_name)
            if isinstance(field, TelegramType):
                field._bind_bot_obj(bot)



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneStr

from .abc import TelegramType


@final
class ForumTopicEdited(TelegramType):
    """This object represents a service message about an edited forum topic.

    See: https://core.telegram.org/bots/api#forumtopicedited
    """

    name: NoneStr = None
    """*Optional.* New name of the topic, if it was edited."""

    icon_custom_emoji_id: NoneStr = None
    """*Optional.* New identifier of the custom emoji shown as the topic icon,
    if it was edited.

    .. note::

        An empty string if the icon was removed.
    """



from __future__ import annotations

__all__ = (
    "parse_command_args",
    "md",
    "create_user_link",
    "create_bot_deeplink",
    "create_group_deeplink",
    "create_channel_deeplink",
    "create_game_deeplink",
    "create_webapp_deeplink",
)

import re
from typing import TYPE_CHECKING
from typing import final

from .enums import MarkdownEntity
from .utils import is_valid_deeplink_parameter
from .utils import is_valid_username

if TYPE_CHECKING:
    from typing import Final

    from .enums import AdminFlag
    from .typing import NoneStr


def parse_command_args(text: str) -> list[str]:
    """Parse command args.

    :param text: Message with command.
    """

    return sp[-1].split() if len(sp := text.split(maxsplit=1)) == 2 else []


@final
class md:
    """Markdown formatting.

    See: https://core.telegram.org/bots/api#markdownv2-style

    .. important::

        MarkdownV2 style only.
    """

    # yapf: disable
    REPL: Final[str] = r"\\\1"
    """:meta private:"""
    PATTERN_TEXT: Final[re.Pattern] = re.compile(f'([{re.escape(r"_*[]()~`>#+-=|{}.!")}])')  # noqa: E501
    """:meta private:"""
    PATTERN_FSTRING: Final[re.Pattern] = re.compile(f'([{re.escape(r"_*[]()~`>#+-=|.!")}])')  # noqa: E501
    """:meta private:"""
    PATTERN_CODE_PRE: Final[re.Pattern] = re.compile(r"([\\\`])")
    """:meta private:"""
    PATTERN_LINK_EMOJI: Final[re.Pattern] = re.compile(r"([\\\\\)])")
    """:meta private:"""
    # yapf: enable

    # yapf: disable
    @staticmethod
    def escape(
        text: str, /, entity: MarkdownEntity | str = MarkdownEntity.TEXT
    ) -> str:
        """Use this method to escape markdown characters in text."""

        # yapf: enable
        entity = str(entity)
        if entity == "fstring":
            return md.PATTERN_FSTRING.sub(md.REPL, text)

        if entity in ["link", "emoji"]:
            return md.PATTERN_LINK_EMOJI.sub(md.REPL, text)

        if entity in ["code", "pre"]:
            return md.PATTERN_CODE_PRE.sub(md.REPL, text)

        return md.PATTERN_TEXT.sub(md.REPL, text)

    @staticmethod
    def bold(text: str, /) -> str:
        return f"*{text}*"

    @staticmethod
    def italic(text: str, /) -> str:
        return f"_{text}_"

    @staticmethod
    def underline(text: str, /) -> str:
        return f"__{text}__"

    @staticmethod
    def strikethrough(text: str, /) -> str:
        return f"~{text}~"

    @staticmethod
    def spoiler(text: str, /) -> str:
        return f"||{text}||"

    @staticmethod
    def url(text: str, url: str, /) -> str:
        return f"[{text}]({url})"

    @staticmethod
    def mention(username: str, user_id: int, /) -> str:
        return f"[{username}](tg://user?id={user_id})"

    @staticmethod
    def emoji(emoji_id: int, placeholder_emoji: str, /) -> str:
        return f"![{placeholder_emoji}](tg://emoji?id={emoji_id})"

    @staticmethod
    def inline(text: str, /) -> str:
        return f"`{text}`"

    @staticmethod
    def code(text: str, /, lang: str = "") -> str:
        return f"```{lang}\n{text}\n```"


def create_user_link(username: str) -> str:
    """Returns link used to share public users, groups and channels

    See: https://core.telegram.org/api/links#public-username-links

    :param username: Public username.
    :raises ValueError: if username contains invalid chars.
    """

    if not is_valid_username(username):
        raise ValueError("Username is invalid.")
    return f"https://t.me/{username}"


def _check_deeplink_parameter(parameter: str) -> None:
    if not is_valid_deeplink_parameter(parameter):
        raise ValueError(
            "`parameter` can contain only following characters: "
            "a-z, A-Z, 0-9, '_', '-'"
        )


def create_bot_deeplink(bot_username: str, parameter: NoneStr = None) -> str:
    """Returns deep link used to link to bots.

    See: https://core.telegram.org/api/links#bot-links

    :param bot_username: Bot username.
    :param parameter: Start parameter, up to 64 base64url characters: if
        provided and the bot_username is indeed a bot, the text input bar should
        be replaced with a Start button (even if the user has already started
        the bot) that should invoke messages.startBot with the appropriate
        parameter once clicked.
    :raises ValueError: if ``bot_username`` or ``parameter`` contains invalid
        chars.
    """

    link = create_user_link(bot_username)
    if parameter:
        _check_deeplink_parameter(parameter)
        link += f"?start={parameter}"

    return link


def create_group_deeplink(
    bot_username: str,
    parameter: NoneStr = None,
    admin: list[AdminFlag] | None = None,
) -> str:
    """Returns deep link used to add bots to groups.

    See: https://core.telegram.org/api/links#groupchannel-bot-links

    :param bot_username: Bot username
    :param parameter: *Optional.* Start parameter, up to 64 base64url
        characters: if provided and the bot_username is indeed a bot,
        messages.startBot with the appropriate parameter should be invoked after
        adding the bot to the group.
    :param admin: *Optional.* A list of identifiers.
    :raises ValueError: if ``bot_username`` or ``parameter`` contains invalid
        chars.
    """

    link = create_user_link(bot_username) + "?startgroup"
    if parameter:
        _check_deeplink_parameter(parameter)
        link += f"={parameter}"

    if admin:
        link += f"&admin={'+'.join(admin)}"

    return link


def create_channel_deeplink(
    bot_username: str,
    admin: list[AdminFlag],
) -> str:
    """Returns deep link used to add bots to channels.

    See: https://core.telegram.org/api/links#groupchannel-bot-links

    :param bot_username: Bot username
    :param admin: A list of identifiers.
    :raises ValueError: if ``bot_username`` contains invalid chars or ``admin``
        is empty.
    """

    if not admin:
        raise ValueError("`admin` can't be empty.")

    return (
        create_user_link(bot_username) +
        f"?startchannel&admin={'+'.join(admin)}"
    )


def create_game_deeplink(bot_username: str, short_name: str) -> str:
    """Returns deep link used to share games.

    See: https://core.telegram.org/api/links#game-links

    :param bot_username: Username of the bot that owns the game.
    :param short_name: Game short name.
    :raises ValueError: if ``bot_username`` contains invalid chars.
    """

    return create_user_link(bot_username) + f"?game={short_name}"


def create_webapp_deeplink(
    bot_username: str,
    short_name: str,
    parameter: NoneStr = None,
) -> str:
    """Returns deep link used to share named bot web apps.

    See: https://core.telegram.org/api/links#named-bot-web-app-links

    :param bot_username: Username of the bot that owns the web app.
    :param short_name: Web app short name.
    :param parameter: *Optional.* `start_param` to pass to
        messages.requestAppWebView.
    :raises ValueError: if ``bot_username`` or ``parameter`` contains invalid
        chars.
    """

    link = create_user_link(bot_username) + f"/{short_name}?startapp"
    if parameter:
        _check_deeplink_parameter(parameter)
        link += f"={parameter}"

    return link



from yatbaf.enums import UpdateType
from yatbaf.filters import Chat
from yatbaf.filters import Command
from yatbaf.handler import Handler


def test_filter(handler_func):
    cmd_filter = Command("cmd")
    chat_filter = Chat()
    handler = Handler(
        handler_func,
        update_type=UpdateType.MESSAGE,
        filters=[cmd_filter, chat_filter],
    )
    assert len(handler._filters) == 2
    assert handler._filters[0] is chat_filter
    assert handler._filters[1] is cmd_filter


def test_filter_sort_false(handler_func):
    cmd_filter = Command("cmd")
    chat_filter = Chat()
    handler = Handler(
        handler_func,
        update_type=UpdateType.MESSAGE,
        filters=[cmd_filter, chat_filter],
        sort_filters=False,
    )
    assert len(handler._filters) == 2
    assert handler._filters[0] is cmd_filter
    assert handler._filters[1] is chat_filter



import pytest


def test_bind_bot_obj(message):
    bot = object()
    message._bind_bot_obj(bot)
    assert message.bot is bot
    assert message.from_.bot is bot
    assert message.chat.bot is bot


def test_bot_not_bound(message):
    with pytest.raises(RuntimeError):
        message.bot



import pytest

from yatbaf.filters import Not
from yatbaf.filters import User

USER_ID = 1212
USER_USERNAME = "test_user"


@pytest.fixture(autouse=True)
def _set_user_attrs(user):
    user.id = USER_ID
    user.username = USER_USERNAME


@pytest.mark.parametrize("username", ("@test_user", "test_user", USER_ID))
def test_true(message, username):
    assert User(username).check(message)


@pytest.mark.parametrize("username", ("@testuser", "test_user1", 23345678))
def test_false(message, username):
    assert not User(username).check(message)


@pytest.mark.parametrize("username", ("@test_user", "test_user", USER_ID))
def test_invert(message, username):
    assert not Not(User(username)).check(message)



import pytest

from yatbaf.enums import UpdateType
from yatbaf.filters import Chat
from yatbaf.handler import Handler


def test_new_hander(handler_func):
    handler = Handler(handler_func, update_type=UpdateType.MESSAGE)
    assert not handler._filters
    assert not handler._middleware
    assert handler._update_type is UpdateType.MESSAGE
    assert str(handler) == "<Handler[type=message]>"
    assert handler._match_fn is all


@pytest.mark.parametrize(
    "objs",
    [
        (
            Handler(f := object(), update_type=UpdateType.MESSAGE),
            Handler(f, update_type=UpdateType.MESSAGE),
        ),
        (
            Handler(f := object(), update_type=UpdateType.MESSAGE),
            Handler(f, update_type=UpdateType.MESSAGE, filters=[Chat()]),
        ),
        (
            Handler(
                f := object(),
                update_type=UpdateType.MESSAGE,
                middleware=[m := object()]
            ),
            Handler(f, update_type=UpdateType.MESSAGE, middleware=[m]),
        ),
    ]
)
def test_eq(objs):
    handler1, handler2 = objs
    assert handler1 == handler2


@pytest.mark.parametrize(
    "h1", [Handler(f := object(), update_type=UpdateType.MESSAGE)]
)
@pytest.mark.parametrize(
    "h2",
    [
        Handler(f, update_type=UpdateType.EDITED_MESSAGE),
        Handler(f, update_type=UpdateType.MESSAGE, middleware=[object()]),
        Handler(object(), update_type=UpdateType.MESSAGE),
    ]
)
def test_not_eq(h1, h2):
    assert h1 != h2



from __future__ import annotations

from typing import final

from .abc import TelegramType
from .reaction_type import ReactionType


@final
class ReactionCount(TelegramType):
    """Represents a reaction added to a message along with the number of times
    it was added.

    See: https://core.telegram.org/bots/api#reactioncount
    """

    type: ReactionType
    """Type of the reaction."""

    total_count: int
    """Number of times the reaction was added."""



from __future__ import annotations

from typing import final

from msgspec import field

from yatbaf.typing import NoneStr

from .abc import TelegramType
from .chat import Chat
from .chat_invite_link import ChatInviteLink
from .user import User

# from datetime import datetime


@final
class ChatJoinRequest(TelegramType):
    """Represents a join request sent to a chat.

    See: https://core.telegram.org/bots/api#chatjoinrequest
    """

    chat: Chat
    """Chat to which the request was sent."""

    from_: User = field(name="from")
    """User that sent the join request."""

    user_chat_id: int
    """Identifier of a private chat with the user who sent the join request."""

    date: int  # TODO: datetime
    """Date the request was sent in Unix time."""

    bio: NoneStr = None
    """*Optional.* Bio of the user."""

    invite_link: ChatInviteLink | None = None
    """*Optional.* Chat invite link that was used by the user to send the join request."""  # noqa: E501

    async def approve(self) -> bool:
        """See :meth:`~yatbaf.bot.Bot.approve_chat_join_request`"""
        return await self.bot.approve_chat_join_request(
            chat_id=self.chat.id,
            user_id=self.from_.id,
        )

    async def decline(self) -> bool:
        """See :meth:`~yatbaf.bot.Bot.decline_chat_join_request`"""
        return await self.bot.decline_chat_join_request(
            chat_id=self.chat.id,
            user_id=self.from_.id,
        )



from __future__ import annotations

__all__ = (
    "Not",
    "Or",
    "And",
)

from typing import TYPE_CHECKING
from typing import Generic
from typing import TypeVar
from typing import final

if TYPE_CHECKING:
    from yatbaf.typing import Filter

T = TypeVar("T")


@final
class Not(Generic[T]):
    """NOT filter.

    Use it to invert filter result::

        @on_message(filters=[Not(User(123))])
        async def handler(message: Message) -> None:
            ...
    """

    __slots__ = (
        "filter",
        "priority",
    )

    def __init__(self, filter: Filter[T], /) -> None:
        """
        :param filter: Filter.
        """

        self.priority = filter.priority
        self.filter = filter

    def check(self, update: T) -> bool:
        return not self.filter.check(update)


@final
class Or(Generic[T]):
    """OR filter."""

    __slots__ = (
        "priority",
        "_filters",
    )

    def __init__(self, *filters: Filter[T]) -> None:
        """
        :param filters: A list of filters.
        """

        self.priority = (min(*filters, key=lambda v: v.priority)).priority
        self._filters = filters

    def check(self, update: T) -> bool:
        return any(f.check(update) for f in self._filters)


@final
class And(Generic[T]):
    """AND filter."""

    __slots__ = (
        "priority",
        "_filters",
    )

    def __init__(self, *filters: Filter[T]) -> None:
        """
        :param filters: A list of filters.
        """

        self.priority = (min(*filters, key=lambda v: v.priority)).priority
        self._filters = filters

    def check(self, update: T) -> bool:
        return all(f.check(update) for f in self._filters)



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class LabeledPrice(TelegramType):
    """This object represents a portion of the price for goods or services.

    See: https://core.telegram.org/bots/api#labeledprice
    """

    label: str
    """Portion label."""

    amount: int
    """Price of the product in the smallest units of the currency.

    .. important::

        Integer, not float/double.
    """



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneInt
from yatbaf.typing import NoneStr

from .abc import TelegramType
from .photo_size import PhotoSize


@final
class Audio(TelegramType):
    """This object represents an audio file to be treated as music by the
    Telegram clients.

    See: https://core.telegram.org/bots/api#audio
    """

    file_id: str
    """
    Identifier for this file, which can be used to download or reuse the file.
    """

    file_unique_id: str
    """Unique identifier for this file, which is supposed to be the same over
    time and for different bots.

    .. warning::

        Can't be used to download or reuse the file.
    """

    duration: int
    """Duration of the audio in seconds as defined by sender."""

    performer: NoneStr = None
    """*Optional.*
    Performer of the audio as defined by sender or by audio tags.
    """

    title: NoneStr = None
    """*Optional.* Title of the audio as defined by sender or by audio tags."""

    file_name: NoneStr = None
    """*Optional.* Original filename as defined by sender."""

    mime_type: NoneStr = None
    """*Optional.* MIME type of the file as defined by sender."""

    file_size: NoneInt = None
    """*Optional.* File size in bytes."""

    thumbnail: PhotoSize | None = None
    """*Optional.*
    Thumbnail of the album cover to which the music file belongs.
    """



from __future__ import annotations

from typing import final

from .abc import TelegramType

# from datetime import datetime


@final
class VideoChatScheduled(TelegramType):
    """This object represents a service message about a video chat scheduled
    in the chat.

    See: https://core.telegram.org/bots/api#videochatscheduled
    """

    start_date: int  # datetime
    """Point in time (Unix timestamp) when the video chat is supposed to be
    started by a chat administrator.
    """



from yatbaf.types import InputMediaAnimation
from yatbaf.types import InputMediaAudio
from yatbaf.types import InputMediaDocument
from yatbaf.types import InputMediaPhoto
from yatbaf.types import InputMediaVideo
from yatbaf.types import InputSticker


def test_attr_is_private(message):
    assert "__usrctx__" not in message.__struct_fields__
    assert "__usrctx__" not in message.__slots__


def test_type_file_fields():
    assert InputSticker.__type_file_fields__ == ("sticker",)
    assert InputMediaPhoto.__type_file_fields__ == ("media",)
    assert InputMediaAudio.__type_file_fields__ == ("media", "thumbnail")
    assert InputMediaVideo.__type_file_fields__ == ("media", "thumbnail")
    assert InputMediaDocument.__type_file_fields__ == ("media", "thumbnail")
    assert InputMediaAnimation.__type_file_fields__ == ("media", "thumbnail")



from __future__ import annotations

__all__ = ("warn_duplicate",)

import warnings
from typing import TYPE_CHECKING

from .exceptions import BotWarning

if TYPE_CHECKING:
    from .abc import AbstractRouter
    from .handler import Handler


def warn_duplicate(
    obj: Handler | AbstractRouter,
    dest: AbstractRouter,
    stacklevel: int = 2,
) -> None:
    warnings.warn(
        f"{obj!r} is a duplicate and was not added to {dest!r}",
        category=BotWarning,
        stacklevel=stacklevel,
    )



from __future__ import annotations

__all__ = (
    "TelegramAPIException",
    "RequestTimeoutError",
    "RequestError",
    "BotException",
    "JSONDecodeError",
    "MethodInvokeError",
    "FileDownloadError",
    "FloodError",
    "ChatMigratedError",
    "InternalError",
    "TokenError",
    "WebhookConflictError",
    "FrozenInstanceError",
    "BotWarning",
)

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .methods.abc import TelegramMethod


class TelegramAPIException(Exception):
    """Base API exception"""


class TokenError(TelegramAPIException):
    """Wrong token. Status 401"""


class WebhookConflictError(TelegramAPIException):
    """Webhook/LongPolling conflict. Status 409"""


class InternalError(TelegramAPIException):
    """Internal server error. Status >= 500"""


class RequestTimeoutError(TelegramAPIException):
    """Request timed out."""

    def __init__(self, method: str, orig: Exception) -> None:
        """
        :param method: Telegram method name.
        :param orig: Original exception.
        """

        super().__init__(f"{method}, {orig}")
        self.method = method
        self.orig = orig


class RequestError(TelegramAPIException):
    """Response status gt 200."""

    def __init__(
        self,
        error_code: int,
        description: str,
    ) -> None:
        """
        :param error_code: HTTP status code.
        :param description: Error description.
        """

        super().__init__(f"{description} [{error_code}]")
        self.error_code = error_code
        self.description = description


class FileDownloadError(RequestError):
    pass


class MethodInvokeError(RequestError):

    def __init__(
        self,
        method: TelegramMethod,
        error_code: int,
        description: str,
    ) -> None:
        """
        :param method: TelegramMethod object.
        :param error_code: HTTP status code.
        :param description: Error description.
        """

        super().__init__(error_code, description)
        self.method = method


class ChatMigratedError(MethodInvokeError):
    """The group has been migrated to a supergroup."""

    def __init__(
        self,
        method: TelegramMethod,
        error_code: int,
        description: str,
        migrate_to_chat_id: int,
    ) -> None:
        """
        :param method: TelegramMethod object.
        :param error_code: HTTP status code.
        :param description: Error description.
        :param migrate_to_chat_id: Supergroup id.
        """

        super().__init__(method, error_code, description)
        self.migrate_to_chat_id = migrate_to_chat_id


class FloodError(MethodInvokeError):
    """Too Many Requests. Status 429"""

    def __init__(
        self,
        method: TelegramMethod,
        error_code: int,
        description: str,
        retry_after: int,
    ) -> None:
        """
        :param method: TelegramMethod object.
        :param error_code: HTTP status code.
        :param description: Error description.
        :param retry_after: The number of seconds left to wait before the
            request can be repeated.
        """

        super().__init__(method, error_code, description)
        self.retry_after = retry_after


class BotException(Exception):
    """Base bot exception"""


class JSONDecodeError(BotException):
    """Decode error."""

    def __init__(self, message: str, raw_content: bytes) -> None:
        """
        :param message: Error message.
        :param raw_content: Response content.
        """

        super().__init__(f"Fail while decoding response content: {message}")
        self.raw_content = raw_content


class InvalidTokenError(BotException):
    """Api token is invalid."""


class FrozenInstanceError(BotException):
    """Object instance is frozen."""


class BotWarning(UserWarning):
    """Base bot warning."""


"""
TelegramAPIException
 ├── TokenError
 ├── WebhookConflictError
 ├── InternalError
 ├── RequestTimeoutError
 └── RequestError
      ├── FileDownloadError
      └── MethodInvokeError
           ├── ChatMigratedError
           └── FloodError
BotException
 ├── JSONDecodeError
 ├── InvalidTokenError
 └── FrozenInstanceError
"""



from __future__ import annotations

from typing import final

from yatbaf.types import File

from .abc import TelegramMethod


@final
class GetFile(TelegramMethod[File]):
    """See :meth:`yatbaf.bot.Bot.get_file`"""

    file_id: str



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import TypeAlias

if TYPE_CHECKING:
    from .input_contact_message_content import InputContactMessageContent
    from .input_invoce_message_content import InputInvoiceMessageContent
    from .input_location_message_content import InputLocationMessageContent
    from .input_text_message_content import InputTextMessageContent
    from .input_venue_message_content import InputVenueMessageContent

# https://core.telegram.org/bots/api#inputmessagecontent
InputMessageContent: TypeAlias = (
    "InputContactMessageContent "
    "| InputInvoiceMessageContent "
    "| InputLocationMessageContent "
    "| InputTextMessageContent "
    "| InputVenueMessageContent"
)



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt
    from yatbaf.typing import NoneStr


@final
class AnswerCallbackQuery(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.answer_callback_query()`"""

    callback_query_id: str
    text: NoneStr = None
    show_alert: NoneBool = None
    url: NoneStr = None
    cache_time: NoneInt = None



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .typing import ErrorHandler

log = logging.getLogger(__name__)


class DefaultErrorHandler:

    def __init__(self, error_handler: ErrorHandler | None = None) -> None:
        self._error_handler = error_handler

    def _log(self, message: str, error: Exception) -> None:
        log.error(f"{message}: {error}")
        log.debug("Exception traceback:", exc_info=error)

    async def on_error(self, error: Exception) -> None:
        if self._error_handler is None:
            self._log("Unexcpected error.", error)
            return

        try:
            await self._error_handler.on_error(error)
        except Exception as error:
            self._log(f"Unexcpected error in {self._error_handler!r}.", error)



from typing import final

from .abc import TelegramType


@final
class CallbackGame(TelegramType):
    """A placeholder, currently holds no information.

    See: https://core.telegram.org/bots/api#callbackgame
    """



from __future__ import annotations

__all__ = (
    "KeyboardButton",
    "KeyboardButtonPollType",
    "KeyboardButtonRequestChat",
    "KeyboardButtonRequestUsers",
)

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramType

if TYPE_CHECKING:
    from yatbaf.enums import PollType
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt

    from .chat_administrator_rights import ChatAdministratorRights
    from .web_app_info import WebAppInfo


@final
class KeyboardButtonPollType(TelegramType):
    """This object represents type of a poll, which is allowed to be created and
    sent when the corresponding button is pressed.

    See: https://core.telegram.org/bots/api#keyboardbuttonpolltype
    """

    type: PollType | None = None
    """*Optional.* If ``quiz`` is passed, the user will be allowed to create
    only polls in the quiz mode. If ``regular`` is passed, only regular polls
    will be allowed. Otherwise, the user will be allowed to create a poll of
    any type.
    """


@final
class KeyboardButtonRequestChat(TelegramType):
    """This object defines the criteria used to request a suitable chat. The
    identifier of the selected chat will be shared with the bot when the
    corresponding button is pressed.

    See: https://core.telegram.org/bots/api#keyboardbuttonrequestchat
    """

    request_id: int
    """Signed 32-bit identifier of the request, which will be received back in
    the :class:`ChatShared <yatbaf.types.chat_shared.ChatShared>` object.

    .. note::

        Must be unique within the message.
    """

    chat_is_channel: bool
    """Pass ``True`` to request a channel chat, pass ``False`` to request a
    group or a supergroup chat.
    """

    chat_is_forum: NoneBool = None
    """*Optional.* Pass ``True`` to request a forum supergroup, pass ``False``
    to request a non-forum chat. If not specified, no additional restrictions
    are applied.
    """

    chat_has_username: NoneBool = None
    """*Optional.* Pass ``True`` to request a supergroup or a channel with a
    username, pass ``False`` to request a chat without a username. If not
    specified, no additional restrictions are applied.
    """

    chat_is_created: NoneBool = None
    """*Optional.* Pass ``True``. to request a chat owned by the user.
    Otherwise, no additional restrictions are applied.
    """

    user_administrator_rights: ChatAdministratorRights | None = None
    """*Optional.* An object listing the required administrator rights of the
    user in the chat. The rights must be a superset of ``bot_administrator_rights``.
    If not specified, no additional restrictions are applied.
    """  # noqa: E501

    bot_administrator_rights: ChatAdministratorRights | None = None
    """*Optional.* An object listing the required administrator rights of the
    bot in the chat. The rights must be a subset of ``user_administrator_rights``.
    If not specified, no additional restrictions are applied.
    """  # noqa: E501

    bot_is_member: NoneBool = None
    """*Optional.* Pass ``True`` to request a chat with the bot as a member.
    Otherwise, no additional restrictions are applied.
    """


@final
class KeyboardButtonRequestUsers(TelegramType):
    """This object defines the criteria used to request a suitable users. The
    identifiers of the selected users will be shared with the bot when the
    corresponding button is pressed.

    See: https://core.telegram.org/bots/api#keyboardbuttonrequestuser
    """

    request_id: int
    """Signed 32-bit identifier of the request, which will be received back in
    the :class:`~yatbaf.types.users_shared.UsersShared` object. Must be unique
    within the message.
    """

    user_is_bot: NoneBool = None
    """*Optional.* Pass ``True`` to request a bot, pass ``False`` to request a
    regular user. If not specified, no additional restrictions are applied.
    """

    user_is_premium: NoneBool = None
    """*Optional.* Pass ``True`` to request a premium user, pass ``False`` to
    request a non-premium user. If not specified, no additional restrictions
    are applied.
    """

    max_quantity: NoneInt = None
    """*Optional.* The maximum number of users to be selected; 1-10.
    Defaults to 1.
    """


@final
class KeyboardButton(TelegramType):
    """This object represents one button of the reply keyboard. For simple text
    buttons, String can be used instead of this object to specify the button
    text.

    .. note::

        The optional fields ``web_app``, ``request_user``, ``request_chat``,
        ``request_contact``, ``request_location``, and ``request_poll`` are
        mutually exclusive.

    See: https://core.telegram.org/bots/api#keyboardbutton
    """

    text: str
    """Text of the button. If none of the optional fields are used, it will be
    sent as a message when the button is pressed.
    """

    request_users: KeyboardButtonRequestUsers | None = None
    """*Optional.* If specified, pressing the button will open a list of
    suitable users. Tapping on any user will send their identifier to the bot
    in a “user_shared” service message. Available in private chats only.
    """

    request_chat: KeyboardButtonRequestChat | None = None
    """*Optional.* If specified, pressing the button will open a list of
    suitable chats. Tapping on a chat will send its identifier to the bot in a
    “chat_shared” service message. Available in private chats only.
    """

    request_contact: NoneBool = None
    """*Optional.* If ``True``, the user's phone number will be sent as a
    contact when the button is pressed. Available in private chats only.
    """

    request_location: NoneBool = None
    """*Optional.* If ``True``, the user's current location will be sent when
    the button is pressed. Available in private chats only.
    """

    request_poll: KeyboardButtonPollType | None = None
    """*Optional.* If specified, the user will be asked to create a poll and
    send it to the bot when the button is pressed. Available in private chats
    only.
    """

    web_app: WebAppInfo | None = None
    """*Optional.* If specified, the described Web App will be launched when
    the button is pressed. The Web App will be able to send a “web_app_data”
    service message. Available in private chats only.
    """



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneStr

from .abc import TelegramType
from .animation import Animation
from .message_entity import MessageEntity
from .photo_size import PhotoSize


@final
class Game(TelegramType):
    """This object represents a game.

    See: https://core.telegram.org/bots/api#game
    """

    title: str
    """Title of the game."""

    description: str
    """Description of the game."""

    photo: list[PhotoSize]
    """Photo that will be displayed in the game message in chats."""

    text: NoneStr = None
    """*Optional.* Brief description of the game or high scores included in the
    game message. Can be automatically edited to include current high scores for
    the game when the bot calls
    :meth:`set_game_score <yatbaf.bot.Bot.set_game_score>`, or manually edited
    using :meth:`edit_message_text <yatbaf.bot.Bot.edit_message_text>`.
    0-4096 characters.
    """

    text_entities: list[MessageEntity] | None = None
    """*Optional.* Special entities that appear in text, such as usernames,
    URLs, bot commands, etc.
    """

    animation: Animation | None = None
    """*Optional.*
    Animation that will be displayed in the game message in chats.
    """



from __future__ import annotations

from typing import final

from .abc import TelegramMethod


@final
class SetStickerKeywords(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.set_sticker_keywords`"""

    sticker: str
    keywords: list[str] | None = None



from typing import final

from .abc import TelegramType


@final
class Story(TelegramType):
    """This object represents a message about a forwarded story in the chat.
    Currently holds no information.

    See: https://core.telegram.org/bots/api#story
    """



from __future__ import annotations

__all__ = (
    "EncryptedCredentials",
    "EncryptedPassportElement",
    "PassportFile",
    "PassportData",
)

# from datetime import datetime

from typing import final

from yatbaf.enums import PassportElement
from yatbaf.typing import NoneStr

from .abc import TelegramType


@final
class EncryptedCredentials(TelegramType):
    """Describes data required for decrypting and authenticating :class:`EncryptedPassportElement`.

    See: https://core.telegram.org/bots/api#encryptedcredentials
    """  # noqa: E501

    data: str
    """Base64-encoded encrypted JSON-serialized data with unique user's payload,
    data hashes and secrets required for :class:`EncryptedPassportElement`
    decryption and authentication.
    """

    hash: str
    """Base64-encoded data hash for data authentication."""

    secret: str
    """Base64-encoded secret, encrypted with the bot's public RSA key, required
    for data decryption.
    """


@final
class PassportFile(TelegramType):
    """This object represents a file uploaded to Telegram Passport.

    .. note::

        Currently all Telegram Passport files are in JPEG format when decrypted
        and don't exceed 10MB.

    See: https://core.telegram.org/bots/api#passportfile
    """

    file_id: str
    """
    Identifier for this file, which can be used to download or reuse the file.
    """

    file_unique_id: str
    """Unique identifier for this file, which is supposed to be the same over
    time and for different bots. Can't be used to download or reuse the file.
    """

    file_size: int
    """File size in bytes."""

    file_date: int  # TODO: datetime?
    """Unix time when the file was uploaded."""


@final
class EncryptedPassportElement(TelegramType):
    """Describes documents or other Telegram Passport elements shared with the
    bot by the user.

    See: https://core.telegram.org/bots/api#encryptedpassportelement
    """

    type: PassportElement
    """Element type."""

    hash: str
    """Base64-encoded element hash for using in :class:`~yatbaf.types.passport_element_error.PassportElementErrorUnspecified`."""  # noqa: E501

    data: NoneStr = None
    """*Optional.* Base64-encoded encrypted Telegram Passport element data
    provided by the user, available for “personal_details”, “passport”,
    “driver_license”, “identity_card”, “internal_passport” and “address” types.
    Can be decrypted and verified using the accompanying
    :class:`EncryptedCredentials`.
    """

    phone_number: NoneStr = None
    """*Optional.* User's verified phone number, available only for
    “phone_number” type.
    """

    email: NoneStr = None
    """
    *Optional.* User's verified email address, available only for “email” type.
    """

    files: list[PassportFile] | None = None
    """*Optional.* Array of encrypted files with documents provided by the user,
    available for “utility_bill”, “bank_statement”, “rental_agreement”,
    “passport_registration” and “temporary_registration” types. Files can be
    decrypted and verified using the accompanying :class:`EncryptedCredentials`.
    """

    front_side: PassportFile | None = None
    """*Optional.* Encrypted file with the front side of the document, provided
    by the user. Available for “passport”, “driver_license”, “identity_card”
    and “internal_passport”. The file can be decrypted and verified using the
    accompanying :class:`EncryptedCredentials`.
    """

    reverse_side: PassportFile | None = None
    """*Optional.* Encrypted file with the reverse side of the document,
    provided by the user. Available for “driver_license” and “identity_card”.
    The file can be decrypted and verified using the accompanying
    :class:`EncryptedCredentials`.
    """

    selfie: PassportFile | None = None
    """*Optional.* Encrypted file with the selfie of the user holding a
    document, provided by the user; available for “passport”, “driver_license”,
    “identity_card” and “internal_passport”. The file can be decrypted and
    verified using the accompanying :class:`EncryptedCredentials`.
    """

    translation: list[PassportFile] | None = None
    """*Optional.* Array of encrypted files with translated versions of
    documents provided by the user. Available if requested for “passport”,
    “driver_license”, “identity_card”, “internal_passport”, “utility_bill”,
    “bank_statement”, “rental_agreement”, “passport_registration” and
    “temporary_registration” types. Files can be decrypted and verified using
    the accompanying :class:`EncryptedCredentials`.
    """


@final
class PassportData(TelegramType):
    """Describes Telegram Passport data shared with the bot by the user.

    See: https://core.telegram.org/bots/api#passportdata
    """

    data: list[EncryptedPassportElement]
    """List with information about documents and other Telegram Passport
    elements that was shared with the bot.
    """

    credentials: EncryptedCredentials
    """Encrypted credentials required to decrypt the data."""



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import final

from yatbaf.types import Message

from .abc import TelegramMethodWithFile

if TYPE_CHECKING:
    from yatbaf.enums import ParseMode
    from yatbaf.types import MessageEntity
    from yatbaf.types import ReplyParameters
    from yatbaf.typing import InputFile
    from yatbaf.typing import NoneBool
    from yatbaf.typing import ReplyMarkup


@final
class SendDocument(TelegramMethodWithFile[Message]):
    """See :meth:`yatbaf.bot.Bot.send_document`"""

    chat_id: str | int
    document: InputFile | str
    message_thread_id: int | None = None
    thumbnail: InputFile | str | None = None
    caption: str | None = None
    parse_mode: ParseMode | None = None
    caption_entities: list[MessageEntity] | None = None
    disable_content_type_detection: NoneBool = None
    disable_notification: NoneBool = None
    protect_content: NoneBool = None
    reply_parameters: ReplyParameters | None = None
    reply_markup: ReplyMarkup | None = None

    __meth_file_fields__: ClassVar[tuple[str, ...]] = (
        "document",
        "thumbnail",
    )



from __future__ import annotations

__all__ = (
    "Handler",
    "on_message",
    "on_edited_message",
    "on_channel_post",
    "on_edited_channel_post",
    "on_message_reaction",
    "on_message_reaction_count",
    "on_inline_query",
    "on_chosen_inline_result",
    "on_callback_query",
    "on_shipping_query",
    "on_pre_checkout_query",
    "on_poll",
    "on_poll_answer",
    "on_my_chat_member",
    "on_chat_member",
    "on_chat_join_request",
    "on_chat_boost",
    "on_removed_chat_boost",
)

from typing import TYPE_CHECKING
from typing import Generic
from typing import TypeAlias
from typing import TypeVar
from typing import cast
from typing import final
from typing import overload

from .enums import UpdateType
from .typing import UpdateT
from .utils import ensure_unique

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Sequence

    from .abc import AbstractRouter
    from .middleware import Middleware
    from .types import CallbackQuery
    from .types import ChatBoostRemoved
    from .types import ChatBoostUpdated
    from .types import ChatJoinRequest
    from .types import ChatMemberUpdated
    from .types import ChosenInlineResult
    from .types import InlineQuery
    from .types import Message
    from .types import MessageReactionCountUpdated
    from .types import MessageReactionUpdated
    from .types import Poll
    from .types import PollAnswer
    from .types import PreCheckoutQuery
    from .types import ShippingQuery
    from .typing import Filter
    from .typing import HandlerCallable
    from .typing import HandlerMiddleware


@final
class Handler(Generic[UpdateT]):
    """Handler object.

    :param fn: Handler function.
    :param update_type: Handler update type. See :class:`~yatbaf.enums.UpdateType`
    :param middleware: *Optional.* A sequence of :class:`~yatbaf.typing.HandlerMiddleware`.
    :param filters: *Optional.* A sequence of :class:`~yatbaf.typing.Filter`.
    :param sort_filters: Pass ``False`` if you want to use your filter order.
        Default ``True``.
    :param any_filter: Pass ``True`` if matching one of the filters is enough.
        Default ``False``.
    """  # noqa: E501

    __slots__ = (
        "_fn",
        "_filters",
        "_middleware",
        "_update_type",
        "_parent",
        "_match_fn",
        "_middleware_stack",
    )

    def __init__(
        self,
        fn: HandlerCallable[UpdateT],
        update_type: UpdateType | str,
        *,
        middleware: Sequence[HandlerMiddleware[UpdateT]] | None = None,
        filters: Sequence[Filter[UpdateT]] | None = None,
        sort_filters: bool = True,
        any_filter: bool = False,
    ) -> None:
        self._fn = fn
        self._update_type = update_type
        self._match_fn = all if not any_filter else any
        self._parent: AbstractRouter | None = None

        self._filters: list[Filter[UpdateT]] = []
        # yapf: disable
        self._middleware: list[HandlerMiddleware[UpdateT]] = ensure_unique(
            middleware or []
        )
        # yapf: enable
        self._register_filters(filters or [], sort_filters)
        self._middleware_stack: HandlerCallable[UpdateT] = self._fn

    def __repr__(self) -> str:
        return f"<Handler[type={self._update_type}]>"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Handler) and (  # yapf: disable
            other is self or (
                # type, callback and middleware is the same - handlers are
                # equal. filters don't matter.
                other._update_type == self._update_type
                and other._fn is self._fn
                and other._middleware == self._middleware
            )
        )

    async def __call__(self, update: UpdateT, /) -> None:
        """Execute middleware and handler.

        :param update: Incoming update.
        """
        await self._middleware_stack(update)

    @property
    def orig(self) -> HandlerCallable[UpdateT]:
        """Original function."""
        return self._fn

    @property
    def update_type(self) -> UpdateType | str:
        """Handler type."""
        return self._update_type

    def _register_filters(
        self, filters: Sequence[Filter[UpdateT]], sort_filters: bool
    ) -> None:
        for func in filters:
            if func not in self._filters:
                self._filters.append(func)
        if sort_filters:
            self._filters.sort(key=lambda v: v.priority)

    def _build_middleware_stack(self) -> HandlerCallable[UpdateT]:
        middleware_stack = self._fn
        router: AbstractRouter | Handler | None = self
        func: HandlerMiddleware[UpdateT] | Middleware[HandlerCallable[UpdateT]]
        while router is not None:
            middleware = cast(
                "list[HandlerMiddleware[UpdateT]]", router._middleware
            )
            for func in reversed(middleware):
                middleware_stack = func(middleware_stack)
            router = router._parent
        return middleware_stack

    def _on_registration(self) -> None:
        self._middleware_stack = self._build_middleware_stack()

    def _match(self, update: UpdateT, /) -> bool:
        return self._match_fn(f.check(update) for f in self._filters)

    def _is_fallback(self) -> bool:
        return not bool(self._filters)


T = TypeVar("T")
Wrapper: TypeAlias = "Callable[[HandlerCallable[T]], Handler[T]]"


@overload
def on_message(__fn: HandlerCallable[Message]) -> Handler[Message]:
    ...


@overload
def on_message(
    *,
    filters: Sequence[Filter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[Message]:
    ...


def on_message(
    __fn: HandlerCallable[Message] | None = None,
    *,
    filters: Sequence[Filter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[Message] | Handler[Message]:
    """Message handler decorator.

    Use this decorator to decorate handler for `message` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[Message]) -> Handler[Message]:
        handler: Handler[Message] = Handler(
            fn=fn,
            update_type=UpdateType.MESSAGE,
            filters=filters,
            middleware=middleware,
            sort_filters=sort_filters,
            any_filter=any_filter,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_edited_message(__fn: HandlerCallable[Message]) -> Handler[Message]:
    ...


@overload
def on_edited_message(
    *,
    filters: Sequence[Filter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[Message]:
    ...


def on_edited_message(
    __fn: HandlerCallable[Message] | None = None,
    *,
    filters: Sequence[Filter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[Message] | Handler[Message]:
    """Edited message handler decorator.

    Use this decorator to decorate handler for `edited_message` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[Message]) -> Handler[Message]:
        handler: Handler[Message] = Handler(
            fn=fn,
            update_type=UpdateType.EDITED_MESSAGE,
            filters=filters,
            middleware=middleware,
            sort_filters=sort_filters,
            any_filter=any_filter,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_message_reaction(
    __fn: HandlerCallable[MessageReactionUpdated]
) -> Handler[MessageReactionUpdated]:
    ...


@overload
def on_message_reaction(  # yapf: disable
    *,
    filters: Sequence[Filter[MessageReactionUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[MessageReactionUpdated]] | None = None,  # noqa: E501
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[MessageReactionUpdated]:
    ...


def on_message_reaction(  # yapf: disable
    __fn: HandlerCallable[MessageReactionUpdated] | None = None,
    *,
    filters: Sequence[Filter[MessageReactionUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[MessageReactionUpdated]] | None = None,  # noqa: E501
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[MessageReactionUpdated] | Handler[MessageReactionUpdated]:
    """Message reaction handler decorator.

    Use this decorator to decorate handler for `message_reaction` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(
        fn: HandlerCallable[MessageReactionUpdated]
    ) -> Handler[MessageReactionUpdated]:
        handler: Handler[MessageReactionUpdated] = Handler(
            fn=fn,
            update_type=UpdateType.MESSAGE_REACTION,
            filters=filters,
            middleware=middleware,
            sort_filters=sort_filters,
            any_filter=any_filter,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_message_reaction_count(
    __fn: HandlerCallable[MessageReactionCountUpdated]
) -> Handler[MessageReactionCountUpdated]:
    ...


@overload
def on_message_reaction_count(  # yapf: disable
    *,
    filters: Sequence[Filter[MessageReactionCountUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[MessageReactionCountUpdated]] | None = None,  # noqa: E501
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[MessageReactionCountUpdated]:
    ...


def on_message_reaction_count(  # yapf: disable
    __fn: HandlerCallable[MessageReactionCountUpdated] | None = None,
    *,
    filters: Sequence[Filter[MessageReactionCountUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[MessageReactionCountUpdated]] | None = None,  # noqa: E501
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[MessageReactionCountUpdated] | Handler[MessageReactionCountUpdated]:  # noqa: E501
    """Message reaction count handler decorator.

    Use this decorator to decorate handler for `message_reaction` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(
        fn: HandlerCallable[MessageReactionCountUpdated]
    ) -> Handler[MessageReactionCountUpdated]:
        handler: Handler[MessageReactionCountUpdated] = Handler(
            fn=fn,
            update_type=UpdateType.MESSAGE_REACTION_COUNT,
            filters=filters,
            middleware=middleware,
            sort_filters=sort_filters,
            any_filter=any_filter,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_channel_post(__fn: HandlerCallable[Message]) -> Handler[Message]:
    ...


@overload
def on_channel_post(
    *,
    filters: Sequence[Filter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[Message]:
    ...


def on_channel_post(
    __fn: HandlerCallable[Message] | None = None,
    *,
    filters: Sequence[Filter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[Message] | Handler[Message]:
    """Channel post handler decorator.

    Use this decorator to decorate handler for `channel_post` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[Message]) -> Handler[Message]:
        handler: Handler[Message] = Handler(
            fn=fn,
            update_type=UpdateType.CHANNEL_POST,
            filters=filters,
            middleware=middleware,
            sort_filters=sort_filters,
            any_filter=any_filter,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_edited_channel_post(__fn: HandlerCallable[Message]) -> Handler[Message]:
    ...


@overload
def on_edited_channel_post(
    *,
    filters: Sequence[Filter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[Message]:
    ...


def on_edited_channel_post(
    __fn: HandlerCallable[Message] | None = None,
    *,
    filters: Sequence[Filter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[Message] | Handler[Message]:
    """Edited channel post handler decorator.

    Use this decorator to decorate handler for `edited_channel_post` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[Message]) -> Handler[Message]:
        handler: Handler[Message] = Handler(
            fn=fn,
            update_type=UpdateType.EDITED_CHANNEL_POST,
            filters=filters,
            middleware=middleware,
            sort_filters=sort_filters,
            any_filter=any_filter,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_inline_query(__fn: HandlerCallable[InlineQuery]) -> Handler[InlineQuery]:
    ...


@overload
def on_inline_query(
    *,
    filters: Sequence[Filter[InlineQuery]] | None = None,
    middleware: Sequence[HandlerMiddleware[InlineQuery]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[InlineQuery]:
    ...


def on_inline_query(
    __fn: HandlerCallable[InlineQuery] | None = None,
    *,
    filters: Sequence[Filter[InlineQuery]] | None = None,
    middleware: Sequence[HandlerMiddleware[InlineQuery]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[InlineQuery] | Handler[InlineQuery]:
    """Inline query handler decorator.

    Use this decorator to decorate handler for `inline_query` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[InlineQuery]) -> Handler[InlineQuery]:
        handler: Handler[InlineQuery] = Handler(
            fn=fn,
            update_type=UpdateType.INLINE_QUERY,
            filters=filters,
            middleware=middleware,
            sort_filters=sort_filters,
            any_filter=any_filter,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_chosen_inline_result(
    __fn: HandlerCallable[ChosenInlineResult]
) -> Handler[ChosenInlineResult]:
    ...


@overload
def on_chosen_inline_result(
    *,
    filters: Sequence[Filter[ChosenInlineResult]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChosenInlineResult]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[ChosenInlineResult]:
    ...


def on_chosen_inline_result(
    __fn: HandlerCallable[ChosenInlineResult] | None = None,
    *,
    filters: Sequence[Filter[ChosenInlineResult]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChosenInlineResult]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[ChosenInlineResult] | Handler[ChosenInlineResult]:
    """Chosen inline result handler decorator.

    Use this decorator to decorate handler for `chosen_inline_result` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(
        fn: HandlerCallable[ChosenInlineResult]
    ) -> Handler[ChosenInlineResult]:
        handler: Handler[ChosenInlineResult] = Handler(
            fn=fn,
            update_type=UpdateType.CHOSEN_INLINE_RESULT,
            filters=filters,
            middleware=middleware,
            sort_filters=sort_filters,
            any_filter=any_filter,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_callback_query(
    __fn: HandlerCallable[CallbackQuery]
) -> Handler[CallbackQuery]:
    ...


@overload
def on_callback_query(
    *,
    filters: Sequence[Filter[CallbackQuery]] | None = None,
    middleware: Sequence[HandlerMiddleware[CallbackQuery]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[CallbackQuery]:
    ...


def on_callback_query(
    __fn: HandlerCallable[CallbackQuery] | None = None,
    *,
    filters: Sequence[Filter[CallbackQuery]] | None = None,
    middleware: Sequence[HandlerMiddleware[CallbackQuery]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[CallbackQuery] | Handler[CallbackQuery]:
    """Callback query handler decorator.

    Use this decorator to decorate handler for `callback_query` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[CallbackQuery]) -> Handler[CallbackQuery]:
        handler: Handler[CallbackQuery] = Handler(
            fn=fn,
            update_type=UpdateType.CALLBACK_QUERY,
            filters=filters,
            middleware=middleware,
            sort_filters=sort_filters,
            any_filter=any_filter,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_shipping_query(
    __fn: HandlerCallable[ShippingQuery]
) -> Handler[ShippingQuery]:
    ...


@overload
def on_shipping_query(
    *,
    filters: Sequence[Filter[ShippingQuery]] | None = None,
    middleware: Sequence[HandlerMiddleware[ShippingQuery]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[ShippingQuery]:
    ...


def on_shipping_query(
    __fn: HandlerCallable[ShippingQuery] | None = None,
    *,
    filters: Sequence[Filter[ShippingQuery]] | None = None,
    middleware: Sequence[HandlerMiddleware[ShippingQuery]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[ShippingQuery] | Handler[ShippingQuery]:
    """Shipping query handler decorator.

    Use this decorator to decorate handler for `shipping_query` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[ShippingQuery]) -> Handler[ShippingQuery]:
        handler: Handler[ShippingQuery] = Handler(
            fn=fn,
            update_type=UpdateType.SHIPPING_QUERY,
            filters=filters,
            middleware=middleware,
            sort_filters=sort_filters,
            any_filter=any_filter,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_pre_checkout_query(
    __fn: HandlerCallable[PreCheckoutQuery]
) -> Handler[PreCheckoutQuery]:
    ...


@overload
def on_pre_checkout_query(
    *,
    filters: Sequence[Filter[PreCheckoutQuery]] | None = None,
    middleware: Sequence[HandlerMiddleware[PreCheckoutQuery]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[PreCheckoutQuery]:
    ...


def on_pre_checkout_query(
    __fn: HandlerCallable[PreCheckoutQuery] | None = None,
    *,
    filters: Sequence[Filter[PreCheckoutQuery]] | None = None,
    middleware: Sequence[HandlerMiddleware[PreCheckoutQuery]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[PreCheckoutQuery] | Handler[PreCheckoutQuery]:
    """Pre-checkout query handler decorator.

    Use this decorator to decorate handler for `pre_checkout_query` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(
        fn: HandlerCallable[PreCheckoutQuery]
    ) -> Handler[PreCheckoutQuery]:
        handler: Handler[PreCheckoutQuery] = Handler(
            fn=fn,
            update_type=UpdateType.PRE_CHECKOUT_QUERY,
            filters=filters,
            middleware=middleware,
            sort_filters=sort_filters,
            any_filter=any_filter,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_poll(__fn: HandlerCallable[Poll]) -> Handler[Poll]:
    ...


@overload
def on_poll(
    *,
    filters: Sequence[Filter[Poll]] | None = None,
    middleware: Sequence[HandlerMiddleware[Poll]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[Poll]:
    ...


def on_poll(
    __fn: HandlerCallable[Poll] | None = None,
    *,
    filters: Sequence[Filter[Poll]] | None = None,
    middleware: Sequence[HandlerMiddleware[Poll]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[Poll] | Handler[Poll]:
    """Poll handler decorator.

    Use this decorator to decorate handler for `poll` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[Poll]) -> Handler[Poll]:
        handler: Handler[Poll] = Handler(
            fn=fn,
            update_type=UpdateType.POLL,
            filters=filters,
            middleware=middleware,
            sort_filters=sort_filters,
            any_filter=any_filter,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_poll_answer(__fn: HandlerCallable[PollAnswer]) -> Handler[PollAnswer]:
    ...


@overload
def on_poll_answer(
    *,
    filters: Sequence[Filter[PollAnswer]] | None = None,
    middleware: Sequence[HandlerMiddleware[PollAnswer]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[PollAnswer]:
    ...


def on_poll_answer(
    __fn: HandlerCallable[PollAnswer] | None = None,
    *,
    filters: Sequence[Filter[PollAnswer]] | None = None,
    middleware: Sequence[HandlerMiddleware[PollAnswer]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[PollAnswer] | Handler[PollAnswer]:
    """Poll answer handler decorator.

    Use this decorator to decorate handler for `poll_answer` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[PollAnswer]) -> Handler[PollAnswer]:
        handler: Handler[PollAnswer] = Handler(
            fn=fn,
            update_type=UpdateType.POLL_ANSWER,
            filters=filters,
            middleware=middleware,
            sort_filters=sort_filters,
            any_filter=any_filter,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_my_chat_member(
    __fn: HandlerCallable[ChatMemberUpdated]
) -> Handler[ChatMemberUpdated]:
    ...


@overload
def on_my_chat_member(
    *,
    filters: Sequence[Filter[ChatMemberUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatMemberUpdated]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[ChatMemberUpdated]:
    ...


def on_my_chat_member(
    __fn: HandlerCallable[ChatMemberUpdated] | None = None,
    *,
    filters: Sequence[Filter[ChatMemberUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatMemberUpdated]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[ChatMemberUpdated] | Handler[ChatMemberUpdated]:
    """My chat member handler decorator.

    Use this decorator to decorate handler for `my_chat_member` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(
        fn: HandlerCallable[ChatMemberUpdated]
    ) -> Handler[ChatMemberUpdated]:
        handler: Handler[ChatMemberUpdated] = Handler(
            fn=fn,
            update_type=UpdateType.MY_CHAT_MEMBER,
            filters=filters,
            middleware=middleware,
            sort_filters=sort_filters,
            any_filter=any_filter,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_chat_member(
    __fn: HandlerCallable[ChatMemberUpdated]
) -> Handler[ChatMemberUpdated]:
    ...


@overload
def on_chat_member(
    *,
    filters: Sequence[Filter[ChatMemberUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatMemberUpdated]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[ChatMemberUpdated]:
    ...


def on_chat_member(
    __fn: HandlerCallable[ChatMemberUpdated] | None = None,
    *,
    filters: Sequence[Filter[ChatMemberUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatMemberUpdated]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[ChatMemberUpdated] | Handler[ChatMemberUpdated]:
    """Chat member handler decorator.

    Use this decorator to decorate handler for `chat_member` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(
        fn: HandlerCallable[ChatMemberUpdated]
    ) -> Handler[ChatMemberUpdated]:
        handler: Handler[ChatMemberUpdated] = Handler(
            fn=fn,
            update_type=UpdateType.CHAT_MEMBER,
            filters=filters,
            middleware=middleware,
            sort_filters=sort_filters,
            any_filter=any_filter,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_chat_join_request(
    __fn: HandlerCallable[ChatJoinRequest]
) -> Handler[ChatJoinRequest]:
    ...


@overload
def on_chat_join_request(
    *,
    filters: Sequence[Filter[ChatJoinRequest]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatJoinRequest]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[ChatJoinRequest]:
    ...


def on_chat_join_request(
    __fn: HandlerCallable[ChatJoinRequest] | None = None,
    *,
    filters: Sequence[Filter[ChatJoinRequest]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatJoinRequest]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[ChatJoinRequest] | Handler[ChatJoinRequest]:
    """Chat join request handler decorator.

    Use this decorator to decorate handler for `chat_join_request` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(
        fn: HandlerCallable[ChatJoinRequest]
    ) -> Handler[ChatJoinRequest]:
        handler: Handler[ChatJoinRequest] = Handler(
            fn=fn,
            update_type=UpdateType.CHAT_JOIN_REQUEST,
            filters=filters,
            middleware=middleware,
            sort_filters=sort_filters,
            any_filter=any_filter,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_chat_boost(
    __fn: HandlerCallable[ChatBoostUpdated]
) -> Handler[ChatBoostUpdated]:
    ...


@overload
def on_chat_boost(
    *,
    filters: Sequence[Filter[ChatBoostUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatBoostUpdated]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[ChatBoostUpdated]:
    ...


def on_chat_boost(
    __fn: HandlerCallable[ChatBoostUpdated] | None = None,
    *,
    filters: Sequence[Filter[ChatBoostUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatBoostUpdated]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[ChatBoostUpdated] | Handler[ChatBoostUpdated]:
    """Chat boost handler decorator.

    Use this decorator to decorate handler for `chat_boost` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(
        fn: HandlerCallable[ChatBoostUpdated]
    ) -> Handler[ChatBoostUpdated]:
        handler: Handler[ChatBoostUpdated] = Handler(
            fn=fn,
            update_type=UpdateType.CHAT_BOOST,
            filters=filters,
            middleware=middleware,
            sort_filters=sort_filters,
            any_filter=any_filter,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_removed_chat_boost(
    __fn: HandlerCallable[ChatBoostRemoved]
) -> Handler[ChatBoostRemoved]:
    ...


@overload
def on_removed_chat_boost(
    *,
    filters: Sequence[Filter[ChatBoostRemoved]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatBoostRemoved]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[ChatBoostRemoved]:
    ...


def on_removed_chat_boost(
    __fn: HandlerCallable[ChatBoostRemoved] | None = None,
    *,
    filters: Sequence[Filter[ChatBoostRemoved]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatBoostRemoved]] | None = None,
    sort_filters: bool = True,
    any_filter: bool = False,
) -> Wrapper[ChatBoostRemoved] | Handler[ChatBoostRemoved]:
    """Removed chat boost handler decorator.

    Use this decorator to decorate handler for `removed_chat_boost` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(
        fn: HandlerCallable[ChatBoostRemoved]
    ) -> Handler[ChatBoostRemoved]:
        handler: Handler[ChatBoostRemoved] = Handler(
            fn=fn,
            update_type=UpdateType.REMOVED_CHAT_BOOST,
            filters=filters,
            middleware=middleware,
            sort_filters=sort_filters,
            any_filter=any_filter,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)



import unittest.mock as mock

from yatbaf.enums import UpdateType
from yatbaf.handler import Handler
from yatbaf.router import OnMessage


def test_middleware(handler_func):
    middleware = mock.Mock()
    handler = Handler(
        handler_func,
        update_type=UpdateType.MESSAGE,
        middleware=[middleware],
    )
    assert middleware in handler._middleware


def test_wrap_middlewares(handler_func):
    middleware = mock.Mock(return_value=(func := mock.AsyncMock()))
    handler = Handler(
        handler_func,
        update_type=UpdateType.MESSAGE,
        middleware=[middleware],
    )
    assert handler._build_middleware_stack() is func


def test_wrap_middlewares_empty(handler_func):
    handler = Handler(
        handler_func,
        update_type=UpdateType.MESSAGE,
    )
    assert handler._build_middleware_stack() is handler_func


def test_wrap_middleware_order(handler_func):
    middleware1 = mock.Mock(return_value=(func1 := mock.AsyncMock()))
    middleware2 = mock.Mock()
    middleware3 = mock.Mock()
    handler = Handler(
        handler_func,
        update_type=UpdateType.MESSAGE,
        middleware=[
            middleware1,
            middleware2,
            middleware3,
        ],
    )
    assert handler._build_middleware_stack() is func1


def test_wrap_middleware_parent(handler_func):
    middleware1 = mock.Mock(return_value=(func1 := mock.AsyncMock()))
    middleware2 = mock.Mock()
    middleware3 = mock.Mock()
    handler = Handler(
        handler_func,
        update_type=UpdateType.MESSAGE,
        middleware=[
            middleware2,
            middleware3,
        ],
    )
    _ = OnMessage(
        middleware=[middleware1],
        handlers=[handler],
    )
    assert handler._build_middleware_stack() is func1



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneInt
from yatbaf.typing import NoneStr

from .abc import TelegramType
from .photo_size import PhotoSize


@final
class Animation(TelegramType):
    """This object represents an animation file
    (GIF or H.264/MPEG-4 AVC video without sound).

    See: https://core.telegram.org/bots/api#animation
    """

    file_id: str
    """identifier for this file, which can be used to download or reuse the file."""  # noqa: E501

    file_unique_id: str
    """Unique identifier for this file, which is supposed to be the same over
    time and for different bots.

    .. warning::

        Can't be used to download or reuse the file.
    """

    width: int
    """Video width as defined by sender."""

    height: int
    """Video height as defined by sender."""

    duration: int
    """Duration of the video in seconds as defined by sender."""

    thumbnail: PhotoSize | None = None
    """*Optional* Animation thumbnail as defined by sender."""

    file_name: NoneStr = None
    """*Optional* Original animation filename as defined by sender."""

    mime_type: NoneStr = None
    """*Optional* MIME type of the file as defined by sender."""

    file_size: NoneInt = None
    """*Optional* File size in bytes. It can be bigger than 2^31 and some
    programming languages may have difficulty/silent defects in interpreting it.
    But it has at most 52 significant bits, so a signed 64-bit integer or
    double-precision float type are safe for storing this value.
    """



import time
import unittest.mock as mock

import pytest

from yatbaf.models import UpdateInfo
from yatbaf.router import OnMessage
from yatbaf.types import Chat
from yatbaf.types import Message
from yatbaf.types import Update
from yatbaf.types import User


@pytest.fixture
def token():
    return "12345678:testtoken"


@pytest.fixture
def handler_func():
    return mock.AsyncMock(return_value=None)


@pytest.fixture
def user():
    return User(
        id=1010,
        username="testuser",
        is_bot=False,
        first_name="Test",
    )


@pytest.fixture
def chat():
    return Chat(
        id=101010,
        type="group",
        username="testchat",
    )


@pytest.fixture
def message(chat, user):
    return Message(
        from_=user,
        chat=chat,
        date=int(time.time()),
        message_id=101010,
    )


@pytest.fixture
def update(message):
    return Update(
        update_id=9999,
        message=message,
    )


@pytest.fixture
def update_info(update):
    return UpdateInfo(
        id=update.update_id,
        content=update.message,
        name="message",
    )


@pytest.fixture
def router():
    return OnMessage()


# yapf: disable
@pytest.fixture
def asyncdef():
    def factory(result=None):
        async def func(*_, **__):  # noqa: U101
            return result
        return func
    return factory
    # yapf: enable



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class InputLocationMessageContent(TelegramType):
    """Represents the content of a location message to be sent as the result of
    an inline query.

    # https://core.telegram.org/bots/api#inputlocationmessagecontent
    """

    latitude: float
    """Latitude of the location in degrees."""

    longitude: float
    """Longitude of the location in degrees."""

    horizontal_accuracy: float | None = None
    """*Optional.* The radius of uncertainty for the location, measured in
    meters; 0-1500.
    """

    live_period: int | None = None
    """*Optional.* Period in seconds for which the location can be updated,
    should be between 60 and 86400.
    """

    heading: int | None = None
    """*Optional.* For live locations, a direction in which the user is moving,
    in degrees. Must be between 1 and 360 if specified.
    """

    proximity_alert_radius: int | None = None
    """*Optional.* For live locations, a maximum distance for proximity alerts
    about approaching another chat member, in meters. Must be between 1 and
    100000 if specified.
    """



from __future__ import annotations

from typing import final

from yatbaf.types import ChatInviteLink

from .abc import TelegramMethod


@final
class RevokeChatInviteLink(TelegramMethod[ChatInviteLink]):
    """See :meth:`yatbaf.bot.Bot.revoke_chat_invite_link`"""

    chat_id: str | int
    invite_link: str



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class MessageAutoDeleteTimerChanged(TelegramType):
    """This object represents a service message about a change in auto-delete
    timer settings.

    See: https://core.telegram.org/bots/api#messageautodeletetimerchanged
    """

    message_auto_delete_time: int
    """New auto-delete time for messages in the chat; in seconds."""



from __future__ import annotations

from typing import final

from .abc import TelegramType
from .chat import Chat
from .reaction_type import ReactionType
from .user import User


@final
class MessageReactionUpdated(TelegramType, kw_only=True):
    """This object represents a change of a reaction on a message performed by
    a user.

    See: https://core.telegram.org/bots/api#messagereactionupdated
    """

    chat: Chat
    """The chat containing the message the user reacted to."""

    message_id: int
    """Unique identifier of the message inside the chat."""

    user: User | None = None
    """*Optional.* The user that changed the reaction, if the user isn't
    anonymous.
    """

    actor_chat: Chat | None = None
    """*Optional.* The chat on behalf of which the reaction was changed, if the
    user is anonymous.
    """

    date: int
    """Date of the change in Unix time."""

    old_reaction: list[ReactionType]
    """Previous list of reaction types that were set by the user."""

    new_reaction: list[ReactionType]
    """New list of reaction types that have been set by the user."""



from yatbaf.middleware import Middleware
from yatbaf.router import OnMessage


def test_init_middleware_handler():
    func = object()
    router = OnMessage(middleware=[func])
    assert Middleware(func, is_handler=True) in router._middleware


def test_init_middleware_router():
    func = object()
    router = OnMessage(middleware=[Middleware(func)])
    assert Middleware(func) in router._middleware


def test_init_middleware_handler_local():
    func = object()
    router = OnMessage(
        middleware=[
            Middleware(
                func,
                is_handler=True,
                is_local=True,
            ),
        ]
    )
    middleware = Middleware(
        func,
        is_handler=True,
        is_local=True,
    )
    assert middleware in router._middleware
    router._on_registration()
    assert middleware not in router._middleware


def test_add_middleware(router):
    func = object()
    router.add_middleware(func)
    assert Middleware(func, is_handler=True) in router._middleware


def test_middleware_handler_decorator(router):

    @router.middleware
    def func(_):  # noqa: U101
        pass

    assert len(router._middleware) == 1
    assert Middleware(func, is_handler=True) in router._middleware


def test_middleware_router_decorator(router):

    @router.middleware(is_handler=False)
    def func(_):  # noqa: U101
        pass

    assert len(router._middleware) == 1
    assert Middleware(func) in router._middleware


def test_middleware_decorator_duplicate(router):

    @router.middleware
    @router.middleware
    def func(_):  # noqa: U101
        pass

    assert len(router._middleware) == 1
    assert Middleware(func, is_handler=True) in router._middleware



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneInt

from ..types import message
from .abc import TelegramType


@final
class GiveawayCompleted(TelegramType):
    """This object represents a service message about the completion of a
    giveaway without public winners.

    See: https://core.telegram.org/bots/api#giveawaycompleted
    """

    winner_count: int
    """Number of winners in the giveaway."""

    unclaimed_prize_count: NoneInt = None
    """*Optional.* Number of undistributed prizes."""

    giveaway_message: message.Message | None = None
    """*Optional.* Message with the giveaway that was completed, if it wasn't
    deleted.
    """



from __future__ import annotations

__all__ = ("Message",)

from typing import TYPE_CHECKING
from typing import final

from msgspec import field

from yatbaf.typing import NoneBool
from yatbaf.typing import NoneInt
from yatbaf.typing import NoneStr

from .abc import TelegramType
from .animation import Animation
from .audio import Audio
from .chat import Chat
from .chat_shared import ChatShared
from .contact import Contact
from .dice import Dice
from .document import Document
from .external_reply_info import ExternalReplyInfo
from .forum_topic_closed import ForumTopicClosed
from .forum_topic_created import ForumTopicCreated
from .forum_topic_edited import ForumTopicEdited
from .forum_topic_reopened import ForumTopicReopened
from .game import Game
from .general_forum_topic_hidden import GeneralForumTopicHidden
from .general_forum_topic_unhidden import GeneralForumTopicUnhidden
from .giveaway import Giveaway
from .giveaway_completed import GiveawayCompleted
from .giveaway_created import GiveawayCreated
from .giveaway_winners import GiveawayWinners
from .inaccessible_message import InaccessibleMessage
from .inline_keyboard_markup import InlineKeyboardMarkup
from .invoice import Invoice
from .link_preview_options import LinkPreviewOptions
from .location import Location
from .message_auto_delete_timer_changed import MessageAutoDeleteTimerChanged
from .message_entity import MessageEntity
from .message_origin import MessageOrigin
from .passport import PassportData
from .photo_size import PhotoSize
from .poll import Poll
from .proximity_alert_triggered import ProximityAlertTriggered
from .reply_parameters import ReplyParameters
from .sticker import Sticker
from .story import Story
from .successful_payment import SuccessfulPayment
from .text_quote import TextQuote
from .user import User
from .users_shared import UsersShared
from .venue import Venue
from .video import Video
from .video_chat_ended import VideoChatEnded
from .video_chat_participants_invited import VideoChatParticipantsInvited
from .video_chat_scheduled import VideoChatScheduled
from .video_chat_started import VideoChatStarted
from .video_note import VideoNote
from .voice import Voice
from .web_app_data import WebAppData
from .write_access_allowed import WriteAccessAllowed

if TYPE_CHECKING:
    from yatbaf.enums import ParseMode
    from yatbaf.enums import PollType
    from yatbaf.types import InputMedia
    from yatbaf.types import InputMediaAudio
    from yatbaf.types import InputMediaDocument
    from yatbaf.types import InputMediaPhoto
    from yatbaf.types import InputMediaVideo
    from yatbaf.types import MessageId
    from yatbaf.typing import InputFile
    from yatbaf.typing import ReplyMarkup

    from .maybe_inaccessible_message import MaybeInaccessibleMessage


@final
class Message(TelegramType, kw_only=True):
    """This object represents a message.

    See: https://core.telegram.org/bots/api#message
    """

    message_id: int
    """Unique message identifier inside this chat."""

    message_thread_id: NoneInt = None
    """*Optional.* Unique identifier of a message thread to which the message
    belongs.

    .. note::

        For supergroups only.
    """

    from_: User | None = field(name="from")
    """*Optional.* Sender of the message; empty for messages sent to channels.

    .. note::

        For backward compatibility, the field contains a fake sender user
        in non-channel chats, if the message was sent on behalf of a chat.
    """

    sender_chat: Chat | None = None
    """*Optional.* Sender of the message, sent on behalf of a chat. For example,
    the channel itself for channel posts, the supergroup itself for messages
    from anonymous group administrators, the linked channel for messages
    automatically forwarded to the discussion group.

    .. note::

        For backward compatibility, the field from contains a fake sender user
        in non-channel chats, if the message was sent on behalf of a chat.
    """

    date: int  # TODO: datetime
    """Date the message was sent in Unix time."""

    chat: Chat
    """Conversation the message belongs to."""

    forward_origin: MessageOrigin | None = None
    """*Optional.* Information about the original message for forwarded messages."""  # noqa: E501

    is_topic_message: NoneBool = None
    """*Optional.* True, if the message is sent to a forum topic."""

    is_automatic_forward: NoneBool = None
    """*Optional.* True, if the message is a channel post that was automatically
    forwarded to the connected discussion group.
    """

    reply_to_message: Message | None = None
    """*Optional.* For replies, the original message.

    .. note::

        :class:`Message` object in this field will not contain further
        ``reply_to_message`` fields even if it itself is a reply.
    """

    external_reply: ExternalReplyInfo | None = None
    """*Optional.* Information about the message that is being replied to, which
    may come from another chat or forum topic.
    """

    quote: TextQuote | None = None
    """*Optional.* For replies that quote part of the original message, the
    quoted part of the message.
    """

    via_bot: User | None = None
    """*Optional.* Bot through which the message was sent."""

    edit_date: NoneInt = None  # TODO: datetime
    """*Optional.* Date the message was last edited in Unix time."""

    has_protected_content: NoneBool = None
    """*Optional.* ``True``, if the message can't be forwarded."""

    media_group_id: NoneStr = None
    """*Optional.* The unique identifier of a media message group this message
    belongs to.
    """

    author_signature: NoneStr = None
    """*Optional.* Signature of the post author for messages in channels, or
    the custom title of an anonymous group administrator.
    """

    text: NoneStr = None
    """*Optional.* For text messages, the actual UTF-8 text of the message."""

    entities: list[MessageEntity] | None = None
    """*Optional.* For text messages, special entities like usernames, URLs,
    bot commands, etc. that appear in the text.
    """

    link_preview_options: LinkPreviewOptions | None = None
    """*Optional.* Options used for link preview generation for the message, if
    it is a text message and link preview options were changed.
    """

    animation: Animation | None = None
    """*Optional.* Message is an animation, information about the animation.

    .. note::

        For backward compatibility, when this field is set, the ``document``
        field will also be set.
    """

    audio: Audio | None = None
    """*Optional.* Message is an audio file, information about the file."""

    document: Document | None = None
    """*Optional.* Message is a general file, information about the file."""

    photo: list[PhotoSize] | None = None
    """*Optional.* Message is a photo, available sizes of the photo."""

    sticker: Sticker | None = None
    """*Optional.* Message is a sticker, information about the sticker."""

    story: Story | None = None
    """*Optional.* Message is a forwarded story"""

    video: Video | None = None
    """*Optional.* Message is a video, information about the video."""

    video_note: VideoNote | None = None
    """
    *Optional.* Message is a video note, information about the video message.
    """

    voice: Voice | None = None
    """*Optional.* Message is a voice message, information about the file."""

    caption: NoneStr = None
    """*Optional.* Caption for the animation, audio, document, photo, video or
    voice.
    """

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* For messages with a caption, special entities like usernames,
    URLs, bot commands, etc. that appear in the caption.
    """

    has_media_spoiler: NoneBool = None
    """*Optional.* True, if the message media is covered by a spoiler animation."""  # noqa: E501

    contact: Contact | None = None
    """*Optional.* Message is a shared contact, information about the contact."""  # noqa: E501

    dice: Dice | None = None
    """*Optional.* Message is a dice with random value."""

    game: Game | None = None
    """*Optional.* Message is a game, information about the game."""

    poll: Poll | None = None
    """*Optional.* Message is a native poll, information about the poll."""

    venue: Venue | None = None
    """*Optional.* Message is a venue, information about the venue.

    .. note::

        For backward compatibility, when this field is set, the ``location``
        field will also be set.
    """

    location: Location | None = None
    """*Optional.* Message is a shared location, information about the location."""  # noqa: E501

    new_chat_members: list[User] | None = None
    """*Optional.* New members that were added to the group or supergroup and
    information about them (the bot itself may be one of these members).
    """

    left_chat_member: User | None = None
    """*Optional.* A member was removed from the group, information about them
    (this member may be the bot itself).
    """

    new_chat_title: NoneStr = None
    """*Optional.* A chat title was changed to this value."""

    new_chat_photo: list[PhotoSize] | None = None
    """*Optional.* A chat photo was change to this value."""

    delete_chat_photo: NoneBool = None
    """*Optional.* Service message: the chat photo was deleted."""

    group_chat_created: NoneBool = None
    """*Optional.* Service message: the group has been created."""

    supergroup_chat_created: NoneBool = None
    """*Optional.* Service message: the supergroup has been created.

    .. note::

        This field can't be received in a message coming through updates,
        because bot can't be a member of a supergroup when it is created.
        It can only be found in ``reply_to_message`` if someone replies to a
        very first message in a directly created supergroup.
    """

    channel_chat_created: NoneBool = None
    """*Optional.* Service message: the channel has been created.

    .. note::

        This field can't be received in a message coming through updates,
        because bot can't be a member of a channel when it is created. It
        can only be found in ``reply_to_message`` if someone replies to a very
        first message in a channel.
    """

    message_auto_delete_timer_changed: MessageAutoDeleteTimerChanged | None = None  # noqa: E501, yapf: disable
    """*Optional.* Service message: auto-delete timer settings changed in the chat."""  # noqa: E501

    migrate_to_chat_id: NoneInt = None
    """*Optional.* The group has been migrated to a supergroup with the
    specified identifier.
    """

    migrate_from_chat_id: NoneInt = None
    """*Optional.* The supergroup has been migrated from a group with the
    specified identifier.
    """

    if TYPE_CHECKING:
        pinned_message: MaybeInaccessibleMessage | None = None
        """*Optional.* Specified message was pinned.

        .. note::

            :class:`Message` object in this field will not contain further
            ``reply_to_message`` fields even if it is itself a reply.
        """
    else:
        pinned_message: Message | None = None

    invoice: Invoice | None = None
    """*Optional.* Message is an invoice for a payment, information about
    the invoice.
    """

    successful_payment: SuccessfulPayment | None = None
    """*Optional.* Message is a service message about a successful payment,
    information about the payment.
    """

    users_shared: UsersShared | None = None
    """*Optional.* Service message: a users where shared with the bot."""

    chat_shared: ChatShared | None = None
    """*Optional.* Service message: a chat was shared with the bot."""

    connected_website: NoneStr = None
    """*Optional.* The domain name of the website on which the user has logged in."""  # noqa: E501

    write_access_allowed: WriteAccessAllowed | None = None
    """*Optional.* Service message: the user allowed the bot added to the
    attachment menu to write messages."""

    passport_data: PassportData | None = None
    """*Optional.* Telegram Passport data."""

    proximity_alert_triggered: ProximityAlertTriggered | None = None
    """*Optional.* Service message. A user in the chat triggered another user's
    proximity alert while sharing Live Location.
    """

    forum_topic_created: ForumTopicCreated | None = None
    """*Optional.* Service message: forum topic created."""

    forum_topic_edited: ForumTopicEdited | None = None
    """*Optional.* Service message: forum topic edited."""

    forum_topic_closed: ForumTopicClosed | None = None
    """*Optional.* Service message: forum topic closed."""

    forum_topic_reopened: ForumTopicReopened | None = None
    """*Optional.* Service message: forum topic reopened."""

    general_forum_topic_hidden: GeneralForumTopicHidden | None = None
    """*Optional.* Service message: the 'General' forum topic hidden."""

    general_forum_topic_unhidden: GeneralForumTopicUnhidden | None = None
    """*Optional.* Service message: the 'General' forum topic unhidden."""

    giveaway_created: GiveawayCreated | None = None
    """*Optional.* Service message: a scheduled giveaway was created."""

    giveaway: Giveaway | None = None
    """*Optional.* The message is a scheduled giveaway message."""

    giveaway_winners: GiveawayWinners | None = None
    """*Optional.* A giveaway with public winners was completed."""

    giveaway_completed: GiveawayCompleted | None = None
    """*Optional.* Service message: a giveaway without public winners was
    completed.
    """

    video_chat_scheduled: VideoChatScheduled | None = None
    """*Optional.* Service message: video chat scheduled."""

    video_chat_started: VideoChatStarted | None = None
    """*Optional.* Service message: video chat started."""

    video_chat_ended: VideoChatEnded | None = None
    """*Optional.* Service message: video chat ended."""

    video_chat_participants_invited: VideoChatParticipantsInvited | None = None
    """*Optional.* Service message: new participants invited to a video chat."""

    web_app_data: WebAppData | None = None
    """*Optional.* Service message: data sent by a Web App."""

    reply_markup: InlineKeyboardMarkup | None = None
    """*Optional.* Inline keyboard attached to the message. ``login_url``
    buttons are represented as ordinary ``url`` buttons.
    """

    def __post_init__(self) -> None:
        super().__post_init__()
        if (pm := self.pinned_message) is not None and pm.date == 0:
            self.pinned_message = InaccessibleMessage(
                chat=pm.chat,
                message_id=pm.message_id,
            )

    async def answer(
        self,
        text: str,
        parse_mode: ParseMode | None = None,
        entities: list[MessageEntity] | None = None,
        link_preview_options: LinkPreviewOptions | None = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send message.

        See: :meth:`~yatbaf.bot.Bot.send_message`
        """
        return await self.bot.send_message(
            chat_id=self.chat.id,
            text=text,
            message_thread_id=self.message_thread_id,
            parse_mode=parse_mode,
            entities=entities,
            link_preview_options=link_preview_options,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_markup=reply_markup,
        )

    async def reply(
        self,
        text: str,
        parse_mode: ParseMode | None = None,
        entities: list[MessageEntity] | None = None,
        link_preview_options: LinkPreviewOptions | None = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Reply to message.

        See: :meth:`~yatbaf.bot.Bot.send_message`
        """
        return await self.bot.send_message(
            chat_id=self.chat.id,
            text=text,
            message_thread_id=self.message_thread_id,
            parse_mode=parse_mode,
            entities=entities,
            link_preview_options=link_preview_options,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=ReplyParameters(message_id=self.message_id),
            reply_markup=reply_markup,
        )

    async def edit(
        self,
        text: str,
        parse_mode: ParseMode | None = None,
        entities: list[MessageEntity] | None = None,
        link_preview_options: LinkPreviewOptions | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
    ) -> Message | bool:
        """Edit message text.

        See: :meth:`~yatbaf.bot.Bot.edit_message_text`
        """
        return await self.bot.edit_message_text(
            text=text,
            chat_id=self.chat.id,
            message_id=self.message_id,
            parse_mode=parse_mode,
            entities=entities,
            link_preview_options=link_preview_options,
            reply_markup=reply_markup,
        )

    async def delete(self) -> bool:
        """Delete message.

        See: :meth:`~yatbaf.bot.Bot.delete_message`
        """
        return await self.bot.delete_message(
            chat_id=self.chat.id,
            message_id=self.message_id,
        )

    async def forward(
        self,
        chat_id: str | int,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
    ) -> Message:
        """Forward message.

        See: :meth:`~yatbaf.bot.Bot.forward_message`
        """
        return await self.bot.forward_message(
            chat_id=chat_id,
            from_chat_id=self.chat.id,
            message_id=self.message_id,
            message_thread_id=self.message_thread_id,
            disable_notification=disable_notification,
            protect_content=protect_content,
        )

    async def copy(
        self,
        chat_id: str | int,
        *,
        caption: NoneStr = None,
        parse_mode: ParseMode | None = None,
        caption_entities: list[MessageEntity] | None = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> MessageId:
        """Copy message.

        See: :meth:`~yatbaf.bot.Bot.copy_message`
        """
        return await self.bot.copy_message(
            chat_id=chat_id,
            from_chat_id=self.chat.id,
            message_id=self.message_id,
            message_thread_id=self.message_thread_id,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=caption_entities,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=reply_parameters,
            reply_markup=reply_markup,
        )

    async def answer_photo(
        self,
        photo: InputFile | str,
        *,
        caption: NoneStr = None,
        parse_mode: ParseMode | None = None,
        caption_entities: list[MessageEntity] | None = None,
        has_spoiler: NoneBool = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send photo.

        See: :meth:`~yatbaf.bot.Bot.send_photo`
        """
        return await self.bot.send_photo(
            chat_id=self.chat.id,
            photo=photo,
            message_thread_id=self.message_thread_id,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=caption_entities,
            has_spoiler=has_spoiler,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_markup=reply_markup,
        )

    async def reply_photo(
        self,
        photo: InputFile | str,
        *,
        caption: NoneStr = None,
        parse_mode: ParseMode | None = None,
        caption_entities: list[MessageEntity] | None = None,
        has_spoiler: NoneBool = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send photo as reply.

        See: :meth:`~yatbaf.bot.Bot.send_photo`
        """
        return await self.bot.send_photo(
            chat_id=self.chat.id,
            photo=photo,
            message_thread_id=self.message_thread_id,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=caption_entities,
            has_spoiler=has_spoiler,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=ReplyParameters(message_id=self.message_id),
            reply_markup=reply_markup,
        )

    async def answer_audio(
        self,
        audio: InputFile | str,
        *,
        caption: NoneStr = None,
        parse_mode: ParseMode | None = None,
        caption_entities: list[MessageEntity] | None = None,
        duration: NoneInt = None,
        performer: NoneStr = None,
        title: NoneStr = None,
        thumbnail: InputFile | str | None = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send audio.

        See: :meth:`~yatbaf.bot.Bot.send_audio`
        """
        return await self.bot.send_audio(
            chat_id=self.chat.id,
            audio=audio,
            message_thread_id=self.message_thread_id,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=caption_entities,
            duration=duration,
            performer=performer,
            title=title,
            thumbnail=thumbnail,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_markup=reply_markup,
        )

    async def reply_with_audio(
        self,
        audio: InputFile | str,
        *,
        caption: NoneStr = None,
        parse_mode: ParseMode | None = None,
        caption_entities: list[MessageEntity] | None = None,
        duration: NoneInt = None,
        performer: NoneStr = None,
        title: NoneStr = None,
        thumbnail: InputFile | str | None = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send audio as reply.

        See: :meth:`~yatbaf.bot.Bot.send_audio`
        """
        return await self.bot.send_audio(
            chat_id=self.chat.id,
            audio=audio,
            message_thread_id=self.message_thread_id,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=caption_entities,
            duration=duration,
            performer=performer,
            title=title,
            thumbnail=thumbnail,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=ReplyParameters(message_id=self.message_id),
            reply_markup=reply_markup,
        )

    async def answer_document(
        self,
        document: InputFile | str,
        *,
        thumbnail: InputFile | str | None = None,
        caption: NoneStr = None,
        parse_mode: ParseMode | None = None,
        caption_entities: list[MessageEntity] | None = None,
        disable_content_type_detection: NoneBool = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send document.

        See: :meth:`~yatbaf.bot.Bot.send_document`
        """
        return await self.bot.send_document(
            chat_id=self.chat.id,
            document=document,
            message_thread_id=self.message_thread_id,
            thumbnail=thumbnail,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=caption_entities,
            disable_content_type_detection=disable_content_type_detection,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_markup=reply_markup,
        )

    async def reply_document(
        self,
        document: InputFile | str,
        *,
        thumbnail: InputFile | str | None = None,
        caption: NoneStr = None,
        parse_mode: ParseMode | None = None,
        caption_entities: list[MessageEntity] | None = None,
        disable_content_type_detection: NoneBool = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send document as reply.

        See: :meth:`~yatbaf.bot.Bot.send_document`
        """
        return await self.bot.send_document(
            chat_id=self.chat.id,
            document=document,
            message_thread_id=self.message_thread_id,
            thumbnail=thumbnail,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=caption_entities,
            disable_content_type_detection=disable_content_type_detection,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=ReplyParameters(message_id=self.message_id),
            reply_markup=reply_markup,
        )

    async def answer_video(
        self,
        video: InputFile | str,
        *,
        duration: NoneInt = None,
        width: NoneInt = None,
        height: NoneInt = None,
        thumbnail: InputFile | str | None = None,
        caption: NoneStr = None,
        parse_mode: ParseMode | None = None,
        caption_entities: list[MessageEntity] | None = None,
        has_spoiler: NoneBool = None,
        supports_streaming: NoneBool = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send video.

        See: :meth:`~yatbaf.bot.Bot.send_video`
        """
        return await self.bot.send_video(
            chat_id=self.chat.id,
            video=video,
            message_thread_id=self.message_thread_id,
            duration=duration,
            width=width,
            height=height,
            thumbnail=thumbnail,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=caption_entities,
            has_spoiler=has_spoiler,
            supports_streaming=supports_streaming,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_markup=reply_markup,
        )

    async def reply_video(
        self,
        video: InputFile | str,
        *,
        duration: NoneInt = None,
        width: NoneInt = None,
        height: NoneInt = None,
        thumbnail: InputFile | str | None = None,
        caption: NoneStr = None,
        parse_mode: ParseMode | None = None,
        caption_entities: list[MessageEntity] | None = None,
        has_spoiler: NoneBool = None,
        supports_streaming: NoneBool = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send video as reply.

        See: :meth:`~yatbaf.bot.Bot.send_video`
        """
        return await self.bot.send_video(
            chat_id=self.chat.id,
            video=video,
            message_thread_id=self.message_thread_id,
            duration=duration,
            width=width,
            height=height,
            thumbnail=thumbnail,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=caption_entities,
            has_spoiler=has_spoiler,
            supports_streaming=supports_streaming,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=ReplyParameters(message_id=self.message_id),
            reply_markup=reply_markup,
        )

    async def answer_animation(
        self,
        animation: InputFile | str,
        *,
        duration: NoneInt = None,
        width: NoneInt = None,
        height: NoneInt = None,
        thumbnail: InputFile | str | None = None,
        caption: NoneStr = None,
        parse_mode: ParseMode | None = None,
        caption_entities: list[MessageEntity] | None = None,
        has_spoiler: NoneBool = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send animation.

        See: :meth:`~yatbaf.bot.Bot.send_animation`
        """
        return await self.bot.send_animation(
            chat_id=self.chat.id,
            animation=animation,
            message_thread_id=self.message_thread_id,
            duration=duration,
            width=width,
            height=height,
            thumbnail=thumbnail,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=caption_entities,
            has_spoiler=has_spoiler,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_markup=reply_markup,
        )

    async def reply_animation(
        self,
        animation: InputFile | str,
        *,
        duration: NoneInt = None,
        width: NoneInt = None,
        height: NoneInt = None,
        thumbnail: InputFile | str | None = None,
        caption: NoneStr = None,
        parse_mode: ParseMode | None = None,
        caption_entities: list[MessageEntity] | None = None,
        has_spoiler: NoneBool = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send annotation as reply.

        See: :meth:`yatbaf.bot.Bot.send_animation`
        """
        return await self.bot.send_animation(
            chat_id=self.chat.id,
            animation=animation,
            message_thread_id=self.message_thread_id,
            duration=duration,
            width=width,
            height=height,
            thumbnail=thumbnail,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=caption_entities,
            has_spoiler=has_spoiler,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=ReplyParameters(message_id=self.message_id),
            reply_markup=reply_markup,
        )

    async def answer_voice(
        self,
        voice: InputFile | str,
        *,
        caption: NoneStr = None,
        parse_mode: ParseMode | None = None,
        caption_entities: list[MessageEntity] | None = None,
        duration: NoneInt = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send voice.

        See: :meth:`yatbaf.bot.Bot.send_voice`
        """
        return await self.bot.send_voice(
            chat_id=self.chat.id,
            voice=voice,
            message_thread_id=self.message_thread_id,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=caption_entities,
            duration=duration,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_markup=reply_markup,
        )

    async def reply_voice(
        self,
        voice: InputFile | str,
        *,
        caption: NoneStr = None,
        parse_mode: ParseMode | None = None,
        caption_entities: list[MessageEntity] | None = None,
        duration: NoneInt = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send voice as reply.

        See: :meth:`yatbaf.bot.Bot.send_voice`
        """
        return await self.bot.send_voice(
            chat_id=self.chat.id,
            voice=voice,
            message_thread_id=self.message_thread_id,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=caption_entities,
            duration=duration,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=ReplyParameters(message_id=self.message_id),
            reply_markup=reply_markup,
        )

    async def answer_video_note(
        self,
        video_note: InputFile | str,
        *,
        duration: NoneInt = None,
        length: NoneInt = None,
        thumbnail: InputFile | str | None = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send video note.

        See: :meth:`~yatbaf.bot.Bot.send_video_note`
        """
        return await self.bot.send_video_note(
            chat_id=self.chat.id,
            video_note=video_note,
            message_thread_id=self.message_thread_id,
            duration=duration,
            length=length,
            thumbnail=thumbnail,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_markup=reply_markup,
        )

    async def reply_video_note(
        self,
        video_note: InputFile | str,
        *,
        duration: NoneInt = None,
        length: NoneInt = None,
        thumbnail: InputFile | str | None = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send video note as reply.

        See: :meth:`~yatbaf.bot.Bot.send_video_note`
        """
        return await self.bot.send_video_note(
            chat_id=self.chat.id,
            video_note=video_note,
            message_thread_id=self.message_thread_id,
            duration=duration,
            length=length,
            thumbnail=thumbnail,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=ReplyParameters(message_id=self.message_id),
            reply_markup=reply_markup,
        )

    # yapf: disable
    async def answer_media_group(
        self,
        media: list[InputMediaAudio | InputMediaDocument | InputMediaPhoto | InputMediaVideo],  # noqa: E501
        *,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
    ) -> list[Message]:
        """Send media group.

        See: :meth:`~yatbaf.bot.Bot.send_media_group`
        """
        # yapf: enable
        return await self.bot.send_media_group(
            chat_id=self.chat.id,
            media=media,
            message_thread_id=self.message_thread_id,
            disable_notification=disable_notification,
            protect_content=protect_content,
        )

    async def reply_media_group(  # yapf: disable
        self,
        media: list[InputMediaAudio | InputMediaDocument | InputMediaPhoto | InputMediaVideo],  # noqa: E501
        *,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
    ) -> list[Message]:
        """Send media group as reply.

        See: :meth:`~yatbaf.bot.Bot.send_media_group`
        """
        return await self.bot.send_media_group(
            chat_id=self.chat.id,
            media=media,
            message_thread_id=self.message_thread_id,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=ReplyParameters(message_id=self.message_id),
        )

    async def answer_location(
        self,
        latitude: float,
        longitude: float,
        *,
        horizontal_accuracy: float | None = None,
        live_period: NoneInt = None,
        heading: NoneInt = None,
        proximity_alert_radius: NoneInt = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send location.

        See: :meth:`~yatbaf.bot.Bot.send_location`
        """
        return await self.bot.send_location(
            chat_id=self.chat.id,
            latitude=latitude,
            longitude=longitude,
            message_thread_id=self.message_thread_id,
            horizontal_accuracy=horizontal_accuracy,
            live_period=live_period,
            heading=heading,
            proximity_alert_radius=proximity_alert_radius,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_markup=reply_markup,
        )

    async def reply_location(
        self,
        latitude: float,
        longitude: float,
        *,
        horizontal_accuracy: float | None = None,
        live_period: NoneInt = None,
        heading: NoneInt = None,
        proximity_alert_radius: NoneInt = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send location as reply.

        See: :meth:`~yatbaf.bot.Bot.send_location`
        """
        return await self.bot.send_location(
            chat_id=self.chat.id,
            latitude=latitude,
            longitude=longitude,
            message_thread_id=self.message_thread_id,
            horizontal_accuracy=horizontal_accuracy,
            live_period=live_period,
            heading=heading,
            proximity_alert_radius=proximity_alert_radius,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=ReplyParameters(message_id=self.message_id),
            reply_markup=reply_markup,
        )

    async def answer_venue(
        self,
        latitude: float,
        longitude: float,
        title: str,
        address: str,
        *,
        foursquare_id: NoneStr = None,
        foursquare_type: NoneStr = None,
        google_place_id: NoneStr = None,
        google_place_type: NoneStr = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send venue.

        See: :meth:`~yatbaf.bot.Bot.send_venue`
        """
        return await self.bot.send_venue(
            chat_id=self.chat.id,
            latitude=latitude,
            longitude=longitude,
            title=title,
            address=address,
            message_thread_id=self.message_thread_id,
            foursquare_id=foursquare_id,
            foursquare_type=foursquare_type,
            google_place_id=google_place_id,
            google_place_type=google_place_type,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_markup=reply_markup,
        )

    async def reply_venue(
        self,
        latitude: float,
        longitude: float,
        title: str,
        address: str,
        *,
        foursquare_id: NoneStr = None,
        foursquare_type: NoneStr = None,
        google_place_id: NoneStr = None,
        google_place_type: NoneStr = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send venue as reply.

        See: :meth:`~yatbaf.bot.Bot.send_venue`
        """
        return await self.bot.send_venue(
            chat_id=self.chat.id,
            latitude=latitude,
            longitude=longitude,
            title=title,
            address=address,
            message_thread_id=self.message_thread_id,
            foursquare_id=foursquare_id,
            foursquare_type=foursquare_type,
            google_place_id=google_place_id,
            google_place_type=google_place_type,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=ReplyParameters(message_id=self.message_id),
            reply_markup=reply_markup,
        )

    async def answer_contact(
        self,
        phone_number: str,
        first_name: str,
        *,
        last_name: NoneStr = None,
        vcard: NoneStr = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send contact.

        See: :meth:`~yatbaf.bot.Bot.send_contact`
        """
        return await self.bot.send_contact(
            chat_id=self.chat.id,
            phone_number=phone_number,
            first_name=first_name,
            message_thread_id=self.message_thread_id,
            last_name=last_name,
            vcard=vcard,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_markup=reply_markup,
        )

    async def reply_contact(
        self,
        phone_number: str,
        first_name: str,
        *,
        last_name: NoneStr = None,
        vcard: NoneStr = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send contact as reply.

        See: :meth:`~yatbaf.bot.Bot.send_contact`
        """
        return await self.bot.send_contact(
            chat_id=self.chat.id,
            phone_number=phone_number,
            first_name=first_name,
            message_thread_id=self.message_thread_id,
            last_name=last_name,
            vcard=vcard,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=ReplyParameters(message_id=self.message_id),
            reply_markup=reply_markup,
        )

    async def answer_poll(
        self,
        question: str,
        options: list[str],
        *,
        is_anonymous: NoneBool = None,
        type: PollType | None = None,
        allows_multiple_answers: NoneBool = None,
        correct_option_id: NoneInt = None,
        explanation: NoneStr = None,
        explanation_parse_mode: ParseMode | None = None,
        explanation_entities: list[MessageEntity] | None = None,
        open_period: NoneInt = None,
        close_date: NoneInt = None,
        is_closed: NoneBool = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send poll.

        See: :meth:`~yatbaf.bot.Bot.send_poll`
        """
        return await self.bot.send_poll(
            chat_id=self.chat.id,
            question=question,
            options=options,
            message_thread_id=self.message_thread_id,
            is_anonymous=is_anonymous,
            type=type,
            allows_multiple_answers=allows_multiple_answers,
            correct_option_id=correct_option_id,
            explanation=explanation,
            explanation_parse_mode=explanation_parse_mode,
            explanation_entities=explanation_entities,
            open_period=open_period,
            close_date=close_date,
            is_closed=is_closed,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_markup=reply_markup,
        )

    async def reply_poll(
        self,
        question: str,
        options: list[str],
        *,
        is_anonymous: NoneBool = None,
        type: PollType | None = None,
        allows_multiple_answers: NoneBool = None,
        correct_option_id: NoneInt = None,
        explanation: NoneStr = None,
        explanation_parse_mode: ParseMode | None = None,
        explanation_entities: list[MessageEntity] | None = None,
        open_period: NoneInt = None,
        close_date: NoneInt = None,
        is_closed: NoneBool = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send poll as reply.

        See: :meth:`~yatbaf.bot.Bot.send_poll`
        """
        return await self.bot.send_poll(
            chat_id=self.chat.id,
            question=question,
            options=options,
            message_thread_id=self.message_thread_id,
            is_anonymous=is_anonymous,
            type=type,
            allows_multiple_answers=allows_multiple_answers,
            correct_option_id=correct_option_id,
            explanation=explanation,
            explanation_parse_mode=explanation_parse_mode,
            explanation_entities=explanation_entities,
            open_period=open_period,
            close_date=close_date,
            is_closed=is_closed,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=ReplyParameters(message_id=self.message_id),
            reply_markup=reply_markup,
        )

    async def stop_poll(
        self, *, reply_markup: InlineKeyboardMarkup | None = None
    ) -> Poll:
        """Stop poll.

        See: :meth:`~yatbaf.bot.Bot.stop_poll`
        """
        return await self.bot.stop_poll(
            chat_id=self.chat.id,
            message_id=self.message_id,
            reply_markup=reply_markup,
        )

    async def answer_dice(
        self,
        *,
        emoji: NoneStr = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send dice.

        See: :meth:`~yatbaf.bot.Bot.send_dice`
        """
        return await self.bot.send_dice(
            chat_id=self.chat.id,
            message_thread_id=self.message_thread_id,
            emoji=emoji,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_markup=reply_markup,
        )

    async def reply_dice(
        self,
        *,
        emoji: NoneStr = None,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        """Send dice as reply.

        See: :meth:`~yatbaf.bot.Bot.send_dice`
        """

        return await self.bot.send_dice(
            chat_id=self.chat.id,
            message_thread_id=self.message_thread_id,
            emoji=emoji,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=ReplyParameters(message_id=self.message_id),
            reply_markup=reply_markup,
        )

    async def anwser_game(
        self,
        game_short_name: str,
        *,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: InlineKeyboardMarkup | None = None,
    ) -> Message:
        """Send game.

        See: :meth:`~yatbaf.bot.Bot.send_game`
        """
        return await self.bot.send_game(
            chat_id=self.chat.id,
            game_short_name=game_short_name,
            message_thread_id=self.message_thread_id,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_markup=reply_markup,
        )

    async def reply_game(
        self,
        game_short_name: str,
        *,
        disable_notification: NoneBool = None,
        protect_content: NoneBool = None,
        reply_markup: InlineKeyboardMarkup | None = None,
    ) -> Message:
        """Send game as reply.

        See: :meth:`~yatbaf.bot.Bot.send_game`
        """
        return await self.bot.send_game(
            chat_id=self.chat.id,
            game_short_name=game_short_name,
            message_thread_id=self.message_thread_id,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=ReplyParameters(message_id=self.message_id),
            reply_markup=reply_markup,
        )

    async def edit_caption(
        self,
        *,
        caption: NoneStr = None,
        parse_mode: ParseMode | None = None,
        caption_entities: list[MessageEntity] | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
    ) -> Message | bool:
        """Edit message caption.

        See: :meth:`~yatbaf.bot.Bot.edit_message_caption`
        """
        return await self.bot.edit_message_caption(
            chat_id=self.chat.id,
            message_id=self.message_id,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=caption_entities,
            reply_markup=reply_markup,
        )

    async def edit_media(
        self,
        media: InputMedia,
        *,
        reply_markup: InlineKeyboardMarkup | None = None,
    ) -> Message | bool:
        """Edit message media.

        See: :meth:`~yatbaf.bot.Bot.edit_message_media`
        """
        return await self.bot.edit_message_media(
            media=media,
            chat_id=self.chat.id,
            message_id=self.message_id,
            reply_markup=reply_markup,
        )

    async def edit_reply_markup(
        self,
        reply_markup: InlineKeyboardMarkup | None = None,
    ) -> Message | bool:
        """Edit message reply markup.

        See: :meth:`~yatbaf.bot.Bot.edit_message_reply_markup`
        """
        return await self.bot.edit_message_reply_markup(
            chat_id=self.chat.id,
            message_id=self.message_id,
            reply_markup=reply_markup,
        )

    async def pin(self, disable_notification: NoneBool = None) -> bool:
        """Pin message.

        See: :meth:`~yatbaf.bot.Bot.pin_chat_message`
        """
        return await self.bot.pin_chat_message(
            chat_id=self.chat.id,
            message_id=self.message_id,
            disable_notification=disable_notification,
        )

    async def unpin(self) -> bool:
        """Unpin message.

        See: :meth:`~yatbaf.bot.Bot.unpin_chat_message`
        """
        return await self.bot.unpin_chat_message(
            chat_id=self.chat.id,
            message_id=self.message_id,
        )



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import SentWebAppMessage

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.types import InlineQueryResult


@final
class AnswerWebAppQuery(TelegramMethod[SentWebAppMessage]):
    """See :meth:`yatbaf.bot.Bot.answer_web_app_query`"""

    web_app_query_id: str
    result: InlineQueryResult



import pytest

from yatbaf.enums import AdminFlag
from yatbaf.helpers import create_bot_deeplink
from yatbaf.helpers import create_channel_deeplink
from yatbaf.helpers import create_game_deeplink
from yatbaf.helpers import create_group_deeplink
from yatbaf.helpers import create_user_link
from yatbaf.helpers import create_webapp_deeplink


@pytest.mark.parametrize("username", ("bot123", "user_bot"))
def test_create_user_link(username):
    assert create_user_link(username) == f"https://t.me/{username}"


@pytest.mark.parametrize("username", ("bot-123", "bt", "@bot"))
def test_create_user_link_error(username):
    with pytest.raises(ValueError):
        create_user_link(username)


@pytest.mark.parametrize("parameter", ("123-payload", None))
def test_create_bot_deeplink(parameter):
    bot_username = "username_bot"
    result = create_bot_deeplink(bot_username, parameter)
    if parameter is not None:
        assert result == f"https://t.me/{bot_username}?start={parameter}"
    else:
        assert result == f"https://t.me/{bot_username}"


def test_create_bot_deeplink_parameter_gt64():
    with pytest.raises(ValueError):
        create_bot_deeplink("botusername", "parameter" * 10)


@pytest.mark.parametrize("parameter", (
    "param@",
    "param&",
    ";;;",
    "foo bar",
))
def test_create_bot_deeplink_parameter_invalid_chars(parameter):
    with pytest.raises(ValueError):
        create_bot_deeplink("botusername", parameter)


def test_create_group_deeplink_empty():
    bot = "botusername"
    assert create_group_deeplink(bot) == f"https://t.me/{bot}?startgroup"


def test_create_group_deeplink_parameter():
    bot = "botusername"
    parameter = "parameter_123"
    assert (
        create_group_deeplink(bot, parameter) ==
        f"https://t.me/{bot}?startgroup={parameter}"
    )


def test_create_group_deeplink_parameter_empty():
    bot = "botusername"
    assert create_group_deeplink(bot, "") == f"https://t.me/{bot}?startgroup"


def test_create_group_deeplink_admin():
    bot = "botusername"
    flags = [AdminFlag.CHANGE_INFO, AdminFlag.EDIT_MESSAGES]
    assert (
        create_group_deeplink(bot, admin=flags) ==
        f"https://t.me/{bot}?startgroup&admin=change_info+edit_messages"
    )


def test_create_group_deeplink_admin_empty():
    bot = "botusername"
    assert (
        create_group_deeplink(bot, admin=[]) == f"https://t.me/{bot}?startgroup"
    )


def test_create_group_deeplink_parameter_admin():
    bot = "botusername"
    parameter = "param_foo-bar"
    flags = [
        AdminFlag.EDIT_MESSAGES,
        AdminFlag.INVITE_USERS,
        AdminFlag.DELETE_MESSAGES,
    ]
    assert (
        create_group_deeplink(bot, parameter, flags) == (
            f"https://t.me/{bot}?startgroup={parameter}"
            "&admin=edit_messages+invite_users+delete_messages"
        )
    )


def test_create_group_deeplink_parameter_error():
    with pytest.raises(ValueError):
        create_group_deeplink("botusername", "parameter#")


def test_create_channel_deeplink():
    bot = "botusername"
    flags = [
        AdminFlag.EDIT_MESSAGES,
        AdminFlag.DELETE_MESSAGES,
    ]
    assert (
        create_channel_deeplink(bot, flags) == (
            f"https://t.me/{bot}?startchannel"
            "&admin=edit_messages+delete_messages"
        )
    )


def test_create_channel_deeplink_admin_empty():
    with pytest.raises(ValueError):
        create_channel_deeplink("botusername", [])


def test_create_game_deeplink():
    bot = "botusername"
    game = "gameshortname"
    assert create_game_deeplink(bot, game) == f"https://t.me/{bot}?game={game}"


def test_create_game_deeplink_error():
    with pytest.raises(ValueError):
        create_game_deeplink("invalidbot#", "gameshortname")


def test_create_webapp_deeplink():
    bot = "botusername"
    webapp = "webappname"
    assert (
        create_webapp_deeplink(bot, webapp) ==
        f"https://t.me/{bot}/{webapp}?startapp"
    )


def test_create_webapp_deeplink_parameter():
    bot = "botusername"
    webapp = "webappname"
    parameter = "parameter"
    assert (
        create_webapp_deeplink(bot, webapp, parameter) ==
        f"https://t.me/{bot}/{webapp}?startapp={parameter}"
    )


def test_create_webapp_deeplink_parameter_empty():
    bot = "botusername"
    webapp = "webappname"
    parameter = ""
    assert (
        create_webapp_deeplink(bot, webapp, parameter) ==
        f"https://t.me/{bot}/{webapp}?startapp"
    )


def test_create_webapp_deeplink_parameter_error():
    with pytest.raises(ValueError):
        create_webapp_deeplink("botusername", "webapp", "invalidparam#")



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneStr

from .abc import TelegramType


@final
class SentWebAppMessage(TelegramType):
    """Describes an inline message sent by a Web App on behalf of a user.

    See: https://core.telegram.org/bots/api#sentwebappmessage
    """

    inline_message_id: NoneStr = None
    """*Optional.* Identifier of the sent inline message. Available only if
    there is an inline keyboard attached to the message.
    """



import unittest.mock as mock

from yatbaf.dispatcher import Dispatcher
from yatbaf.middleware import Middleware


def test_middleware(token):
    fn = mock.Mock(return_value=(mdwl := mock.Mock))
    dispatcher = Dispatcher(token, middleware=[Middleware(fn)])
    assert not dispatcher._middleware
    assert dispatcher._middleware_stack is mdwl


def test_middleware_order(token):
    fn1 = mock.Mock(return_value=(mdwl1 := mock.Mock))
    fn2 = mock.Mock()
    dispatcher = Dispatcher(
        token, middleware=[
            Middleware(fn1),
            Middleware(fn2),
        ]
    )
    assert not dispatcher._middleware
    assert dispatcher._middleware_stack is mdwl1



import pytest

from yatbaf.enums import UpdateType
from yatbaf.handler import Handler


@pytest.mark.asyncio
async def test_orig_func(handler_func, message):
    handler = Handler(
        handler_func,
        update_type=UpdateType.MESSAGE,
    )
    await handler.orig(message)
    handler_func.assert_awaited_once_with(message)



from __future__ import annotations

from typing import final

from .abc import TelegramType
from .chat import Chat
from .reaction_count import ReactionCount


@final
class MessageReactionCountUpdated(TelegramType):
    """This object represents reaction changes on a message with anonymous
    reactions.

    See: https://core.telegram.org/bots/api#messagereactioncountupdated
    """

    chat: Chat
    """The chat containing the message."""

    message_id: int
    """Unique message identifier inside the chat."""

    date: int
    """Date of the change in Unix time."""

    reactions: list[ReactionCount]
    """List of reactions that are present on the message."""



from __future__ import annotations

from typing import final

from .abc import TelegramMethod


@final
class SetChatAdministratorCustomTitle(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.set_chat_administrator_custom_title`"""

    chat_id: str | int
    user_id: int
    custom_title: str



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.types import BotCommand
    from yatbaf.types import BotCommandScope
    from yatbaf.typing import NoneStr


@final
class SetMyCommands(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.set_my_commands`"""

    commands: list[BotCommand]
    scope: BotCommandScope | None = None
    language_code: NoneStr = None



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class BotName(TelegramType):
    """This object represents the bot's name.

    See: https://core.telegram.org/bots/api#botname
    """

    name: str
    """The bot's name."""



from __future__ import annotations

__all__ = ("Content",)

from typing import TYPE_CHECKING
from typing import Final
from typing import final

from yatbaf.enums import ContentType

if TYPE_CHECKING:
    from yatbaf.types import Message


@final
class Content:
    """Content filter.

    Use it to filter message by content::

        @on_message(filters=[Content("photo")])
        async def process_photo(message: Message) -> None:
            ...

        @on_message(filters=[Content(ContentType.DOCUMENT)])
        async def process_document(message: Message) -> None:
            ...

    See :class:`~yatbaf.enums.ContentType`.
    """

    __slots__ = (
        "content",
        "priority",
    )

    def __init__(
        self,
        *content: ContentType | str,
        priority: int = 300,
    ) -> None:
        """
        :param content: Content type.
        :param priority: Filter priority. Default 300.
        :raise ValueError: If ``content`` is empty or wrong type was passed.
        """
        if not content:
            raise ValueError("You must pass at least one type.")
        self.content = frozenset([ContentType(c) for c in content])
        self.priority = priority

    def check(self, update: Message) -> bool:
        for c in self.content:
            if getattr(update, c) is not None:
                return True
        return False


media_content: Final[Content] = Content(
    ContentType.AUDIO,
    ContentType.VIDEO,
    ContentType.VIDEO_NOTE,
    ContentType.VOICE,
    ContentType.PHOTO,
    ContentType.ANIMATION,
    ContentType.STICKER,
    ContentType.DOCUMENT,
)
"""Media content filter.

Use it to filter message with: ``animation``, ``audio``, ``document``,
``photo``, ``sticker``, ``video``, ``video note`` or ``voice``.

Usage::

    @on_message(filters=[media_content])
    async def callback(message: Message) -> None:
        ...
"""



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class WebAppInfo(TelegramType):
    """Describes a Web App.

    See: https://core.telegram.org/bots/api#webappinfo
    """

    url: str
    """An HTTPS URL of a Web App to be opened with additional data as specified
    in Initializing Web Apps.
    """



from __future__ import annotations

from typing import final

from .abc import TelegramType
from .location import Location


@final
class ChatLocation(TelegramType):
    """Represents a location to which a chat is connected.

    See: https://core.telegram.org/bots/api#chatlocation
    """

    location: Location
    """The location to which the supergroup is connected.

    .. note::

        Can't be a live location.
    """

    address: str
    """Location address; 1-64 characters, as defined by the chat owner."""



from __future__ import annotations

from typing import final

from .abc import TelegramMethod


@final
class SetStickerSetTitle(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.set_sticker_set_title`"""

    name: str
    title: str



import pytest

from yatbaf.enums import ContentType
from yatbaf.filters import Content


def test_empty():
    with pytest.raises(ValueError):
        Content()


def test_wrong_type():
    with pytest.raises(ValueError):
        Content("typo")


@pytest.mark.parametrize("content", ["text", ContentType.TEXT])
def test_true(message, content):
    message.text = "123"
    assert Content(content).check(message)


@pytest.mark.parametrize("content", ["document", ContentType.DOCUMENT])
def test_filter_content_false(message, content):
    message.text = "123"
    message.document = None
    assert not Content(content).check(message)
